# mypy: disable-error-code="no-untyped-def"

from pathlib import Path
from typing import Never
from typing import assert_never
from typing import cast

import pytest
from rdflib import BNode
from rdflib import URIRef

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.cli.args import ValidateDataConfig
from dsp_tools.cli.args import ValidationSeverity
from dsp_tools.clients.authentication_client import AuthenticationClient
from dsp_tools.clients.authentication_client_live import AuthenticationClientLive
from dsp_tools.commands.validate_data.models.input_problems import ProblemType
from dsp_tools.commands.validate_data.models.input_problems import SortedProblems
from dsp_tools.commands.validate_data.models.input_problems import UnknownClassesInData
from dsp_tools.commands.validate_data.models.validation import DetailBaseInfo
from dsp_tools.commands.validate_data.models.validation import ValidationReportGraphs
from dsp_tools.commands.validate_data.process_validation_report.get_user_validation_message import sort_user_problems
from dsp_tools.commands.validate_data.process_validation_report.query_validation_result import (
    _extract_base_info_of_resource_results,
)
from dsp_tools.commands.validate_data.process_validation_report.query_validation_result import reformat_validation_graph
from dsp_tools.commands.validate_data.shacl_cli_validator import ShaclCliValidator
from dsp_tools.commands.validate_data.validate_data import _get_validation_status
from dsp_tools.commands.validate_data.validate_data import _validate_data
from dsp_tools.commands.validate_data.validation.get_validation_report import get_validation_report
from test.e2e.commands.validate_data.util import prepare_data_for_validation_from_file

# ruff: noqa: ARG001 Unused function argument


CONFIG = ValidateDataConfig(
    xml_file=Path(),
    save_graph_dir=None,
    severity=ValidationSeverity.INFO,
    ignore_duplicate_files_warning=False,
    is_on_prod_server=False,
)


@pytest.fixture(scope="module")
def authentication(creds: ServerCredentials) -> AuthenticationClient:
    auth = AuthenticationClientLive(server=creds.server, email=creds.user, password=creds.password)
    return auth


@pytest.fixture(scope="module")
def content_violation_report(
    create_generic_project, authentication, shacl_validator: ShaclCliValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/content_violation.xml")
    graphs, _ = prepare_data_for_validation_from_file(file, authentication, CONFIG.ignore_duplicate_files_warning)
    return get_validation_report(graphs, shacl_validator)


@pytest.fixture(scope="module")
def every_violation_combination_once_report(
    create_generic_project, authentication, shacl_validator: ShaclCliValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/every_violation_combination_once.xml")
    graphs, _ = prepare_data_for_validation_from_file(file, authentication, CONFIG.ignore_duplicate_files_warning)
    return get_validation_report(graphs, shacl_validator)


class TestWithReportGraphs:
    def test_extract_identifiers_of_resource_results(
        self, every_violation_combination_once_report: ValidationReportGraphs
    ) -> None:
        report_and_onto = (
            every_violation_combination_once_report.validation_graph
            + every_violation_combination_once_report.onto_graph
        )
        data_and_onto = (
            every_violation_combination_once_report.data_graph + every_violation_combination_once_report.onto_graph
        )
        result = _extract_base_info_of_resource_results(report_and_onto, data_and_onto)
        result_sorted = sorted(result, key=lambda x: str(x.focus_node_iri))
        expected_iris = [
            (URIRef("http://data/bitstream_no_legal_info"), None),
            (URIRef("http://data/bitstream_no_legal_info"), None),
            (URIRef("http://data/bitstream_no_legal_info"), None),
            (URIRef("http://data/duplicate_still_image_1"), None),
            (URIRef("http://data/duplicate_still_image_2"), None),
            (URIRef("http://data/empty_label"), None),
            (URIRef("http://data/geoname_not_number"), None),
            (URIRef("http://data/id_card_one"), None),
            (URIRef("http://data/id_closed_constraint"), None),
            (URIRef("http://data/id_max_card"), None),
            (URIRef("http://data/id_missing_file_value"), None),
            (URIRef("http://data/identical_values"), None),
            (URIRef("http://data/image_no_legal_info"), None),
            (URIRef("http://data/image_no_legal_info"), None),
            (URIRef("http://data/image_no_legal_info"), None),
            (URIRef("http://data/inexistent_license_iri"), None),
            (URIRef("http://data/label_with_newline"), None),
            (URIRef("http://data/link_target_non_existent"), BNode),
            (URIRef("http://data/link_target_wrong_class"), BNode),
            (URIRef("http://data/list_node_non_existent"), BNode),
            (URIRef("http://data/missing_seqnum"), None),
            (URIRef("http://data/richtext_standoff_link_nonexistent"), None),
            (URIRef("http://data/simpletext_wrong_value_type"), BNode),
            (URIRef("http://data/uri_wrong_value_type"), None),
            (URIRef("http://data/video_segment_start_larger_than_end"), None),
            (URIRef("http://data/video_segment_wrong_bounds"), None),
            (URIRef("http://data/video_segment_wrong_bounds"), None),
        ]
        assert len(result) == len(expected_iris)
        for result_info, expected_iri in zip(result_sorted, expected_iris):
            assert result_info.focus_node_iri == expected_iri[0]
            if expected_iri[1] is None:
                assert not result_info.detail
            else:
                detail_base_info = result_info.detail
                assert isinstance(detail_base_info, DetailBaseInfo)
                assert isinstance(detail_base_info.detail_bn, expected_iri[1])

    def test_reformat_content_violation(self, content_violation_report: ValidationReportGraphs) -> None:
        assert not content_violation_report.conforms
        expected_info_tuples = [
            (
                "comment_on_value_empty",
                "onto:testUriValue",
                "The comment on the value must be a non-empty string",
            ),
            (
                "comment_on_value_whitespace",
                "onto:testUriValue",
                "The comment on the value must be a non-empty string",
            ),
            ("empty_label", "rdfs:label", "The label must be a non-empty string without newlines."),
            ("empty_text_rich", "onto:testRichtext", "The value must be a non-empty string"),
            ("empty_text_simple", "onto:testTextarea", "The value must be a non-empty string"),
            ("geoname_not_number", "onto:testGeoname", "The value must be a valid geoname code"),
            ("label_with_newline", "rdfs:label", "The label must be a non-empty string without newlines."),
            ("link_target_non_existent", "onto:testHasLinkTo", "other"),
            ("link_target_wrong_class", "onto:testHasLinkToCardOneResource", "id_9_target"),
            (
                "list_name_attrib_empty",
                "onto:testListProp",
                (
                    "A valid node from the list 'firstList' must be used with this property "
                    "(input displayed in format 'listName / NodeName')."
                ),
            ),
            (
                "list_name_non_existent",
                "onto:testListProp",
                (
                    "A valid node from the list 'firstList' must be used with this property "
                    "(input displayed in format 'listName / NodeName')."
                ),
            ),
            (
                "list_node_non_existent",
                "onto:testListProp",
                (
                    "A valid node from the list 'firstList' must be used with this property "
                    "(input displayed in format 'listName / NodeName')."
                ),
            ),
            (
                "richtext_standoff_link_nonexistent",
                "hasStandoffLinkTo",
                "A stand-off link must target an existing resource.",
            ),
            (
                "simple_text_with_newlines",
                "onto:testSimpleText",
                "The value must be a non-empty string without newlines.",
            ),
            ("text_only_whitespace_simple", "onto:testTextarea", "The value must be a non-empty string"),
            (
                "wrong_list_used",
                "onto:testListProp",
                (
                    "A valid node from the list 'firstList' must be used with this property "
                    "(input displayed in format 'listName / NodeName')."
                ),
            ),
        ]
        # check the reformatting result
        result = reformat_validation_graph(content_violation_report)
        assert not result.unexpected_results
        assert len(result.problems) == len(expected_info_tuples)
        # check the result of the sorted problems
        sorted_problems = sort_user_problems(result)
        assert len(sorted_problems.unique_violations) == len(expected_info_tuples)
        assert not sorted_problems.user_warnings
        assert not sorted_problems.user_info
        assert not sorted_problems.unexpected_shacl_validation_components
        alphabetically_sorted = sorted(sorted_problems.unique_violations, key=lambda x: str(x.res_id))
        for one_result, expected_info in zip(alphabetically_sorted, expected_info_tuples):
            assert one_result.res_id == expected_info[0]
            assert one_result.prop_name == expected_info[1]
            # depending on the ProblemType, the string we want to test against is in a different parameter
            if one_result.problem_type == ProblemType.INPUT_REGEX:
                assert one_result.expected == expected_info[2]
            elif one_result.problem_type == ProblemType.GENERIC:
                assert one_result.message == expected_info[2]
            elif one_result.problem_type == ProblemType.LINK_TARGET_TYPE_MISMATCH:
                assert one_result.input_value == expected_info[2]
            elif one_result.problem_type == ProblemType.INEXISTENT_LINKED_RESOURCE:
                assert one_result.input_value == expected_info[2]
            else:
                nev: Never = cast(Never, one_result.problem_type)
                assert_never(nev)
        assert not _get_validation_status(sorted_problems, is_on_prod=True)
        assert not _get_validation_status(sorted_problems, is_on_prod=False)

    def test_reformat_every_constraint_once(
        self, every_violation_combination_once_report: ValidationReportGraphs
    ) -> None:
        assert not every_violation_combination_once_report.conforms
        expected_violations = [
            ("empty_label", ProblemType.INPUT_REGEX),
            ("geoname_not_number", ProblemType.INPUT_REGEX),
            ("id_card_one", ProblemType.MIN_CARD),
            ("id_closed_constraint", ProblemType.NON_EXISTING_CARD),
            ("id_max_card", ProblemType.MAX_CARD),
            ("id_missing_file_value", ProblemType.FILE_VALUE),
            ("identical_values", ProblemType.DUPLICATE_VALUE),
            ("inexistent_license_iri", ProblemType.GENERIC),
            ("label_with_newline", ProblemType.GENERIC),
            ("link_target_non_existent", ProblemType.INEXISTENT_LINKED_RESOURCE),
            ("link_target_wrong_class", ProblemType.LINK_TARGET_TYPE_MISMATCH),
            ("list_node_non_existent", ProblemType.GENERIC),
            ("missing_seqnum", ProblemType.GENERIC),
            ("richtext_standoff_link_nonexistent", ProblemType.GENERIC),
            ("simpletext_wrong_value_type", ProblemType.VALUE_TYPE_MISMATCH),
            ("uri_wrong_value_type", ProblemType.VALUE_TYPE_MISMATCH),
            ("video_segment_start_larger_than_end", ProblemType.GENERIC),
            ("video_segment_wrong_bounds", ProblemType.GENERIC),  # once for start that is less than zero
            ("video_segment_wrong_bounds", ProblemType.GENERIC),  # once for the end that is zero
        ]
        expected_warnings = [
            ("bitstream_no_legal_info", ProblemType.GENERIC),
            ("bitstream_no_legal_info", ProblemType.GENERIC),
            ("bitstream_no_legal_info", ProblemType.GENERIC),
            ("image_no_legal_info", ProblemType.GENERIC),
            ("image_no_legal_info", ProblemType.GENERIC),
            ("image_no_legal_info", ProblemType.GENERIC),
        ]
        expected_info = [
            ("duplicate_still_image_1", ProblemType.FILE_DUPLICATE),
            ("duplicate_still_image_2", ProblemType.FILE_DUPLICATE),
        ]
        result = reformat_validation_graph(every_violation_combination_once_report)
        sorted_problems = sort_user_problems(result)
        alphabetically_sorted_violations = sorted(sorted_problems.unique_violations, key=lambda x: str(x.res_id))
        alphabetically_sorted_warnings = sorted(sorted_problems.user_warnings, key=lambda x: str(x.res_id))
        alphabetically_sorted_info = sorted(sorted_problems.user_info, key=lambda x: str(x.res_id))
        assert len(sorted_problems.unique_violations) == len(expected_violations)
        assert len(sorted_problems.user_warnings) == len(expected_warnings)
        assert len(sorted_problems.user_info) == len(expected_info)
        assert not sorted_problems.unexpected_shacl_validation_components
        assert not result.unexpected_results
        for one_result, expected in zip(alphabetically_sorted_violations, expected_violations):
            assert one_result.res_id == expected[0]
            assert one_result.problem_type == expected[1]
        for one_result, expected in zip(alphabetically_sorted_warnings, expected_warnings):
            assert one_result.problem_type == expected[1]
            assert one_result.res_id == expected[0]
        for one_result, expected in zip(alphabetically_sorted_info, expected_info):
            assert one_result.problem_type == expected[1]
            assert one_result.res_id == expected[0]
        assert not _get_validation_status(sorted_problems, is_on_prod=True)
        assert not _get_validation_status(sorted_problems, is_on_prod=False)


@pytest.mark.usefixtures("create_generic_project")
def test_check_for_unknown_resource_classes(authentication) -> None:
    file = Path("testdata/validate-data/generic/unknown_classes.xml")
    graphs, used_iris = prepare_data_for_validation_from_file(
        file, authentication, CONFIG.ignore_duplicate_files_warning
    )
    result = _validate_data(graphs, used_iris, CONFIG)
    assert not result.no_problems
    problems = result.problems
    assert isinstance(problems, UnknownClassesInData)
    expected = {"onto:NonExisting", "unknown:ClassWithEverything", "unknownClass"}
    assert problems.unknown_classes == expected


@pytest.mark.usefixtures("create_generic_project")
def test_reformat_cardinality_violation(authentication) -> None:
    file = Path("testdata/validate-data/generic/cardinality_violation.xml")
    graphs, used_iris = prepare_data_for_validation_from_file(
        file, authentication, CONFIG.ignore_duplicate_files_warning
    )
    result = _validate_data(graphs, used_iris, CONFIG)
    assert not result.no_problems
    expected_info_tuples = [
        ("id_card_one", ProblemType.MIN_CARD),
        ("id_closed_constraint", ProblemType.NON_EXISTING_CARD),
        ("id_max_card", ProblemType.MAX_CARD),
        ("id_min_card", ProblemType.MIN_CARD),
        ("super_prop_no_card", ProblemType.NON_EXISTING_CARD),
    ]
    sorted_problems = result.problems
    assert isinstance(sorted_problems, SortedProblems)
    assert len(sorted_problems.unique_violations) == len(expected_info_tuples)
    assert not sorted_problems.user_warnings
    assert not sorted_problems.user_info
    assert not sorted_problems.unexpected_shacl_validation_components
    alphabetically_sorted = sorted(sorted_problems.unique_violations, key=lambda x: str(x.res_id))
    for one_result, expected_info in zip(alphabetically_sorted, expected_info_tuples):
        assert one_result.res_id == expected_info[0]
        assert one_result.problem_type == expected_info[1]
    assert not _get_validation_status(sorted_problems, is_on_prod=True)
    assert not _get_validation_status(sorted_problems, is_on_prod=False)


@pytest.mark.usefixtures("create_generic_project")
def test_reformat_value_type_violation(authentication) -> None:
    file = Path("testdata/validate-data/generic/value_type_violation.xml")
    graphs, used_iris = prepare_data_for_validation_from_file(
        file, authentication, CONFIG.ignore_duplicate_files_warning
    )
    result = _validate_data(graphs, used_iris, CONFIG)
    assert not result.no_problems
    expected_info_tuples = [
        ("bool_wrong_value_type", "This property requires a BooleanValue", "onto:testBoolean"),
        ("color_wrong_value_type", "This property requires a ColorValue", "onto:testColor"),
        ("date_wrong_value_type", "This property requires a DateValue", "onto:testSubDate1"),
        ("decimal_wrong_value_type", "This property requires a DecimalValue", "onto:testDecimalSimpleText"),
        ("geoname_wrong_value_type", "This property requires a GeonameValue", "onto:testGeoname"),
        ("integer_wrong_value_type", "This property requires a IntValue", "onto:testIntegerSimpleText"),
        ("is_date_should_be_simpletext", "This property requires a TextValue", "onto:testTextarea"),
        ("is_link_should_be_text", "TextValue without formatting", "onto:testTextarea"),
        ("is_text_should_be_integer", "This property requires a IntValue", "onto:testIntegerSpinbox"),
        ("link_wrong_value_type", "This property requires a LinkValue", "onto:testHasLinkTo"),
        ("list_wrong_value_type", "This property requires a ListValue", "onto:testListProp"),
        ("richtext_wrong_value_type", "TextValue with formatting", "onto:testRichtext"),
        ("simpletext_wrong_value_type", "TextValue without formatting", "onto:testTextarea"),
        ("time_wrong_value_type", "This property requires a TimeValue", "onto:testTimeValue"),
        ("uri_wrong_value_type", "This property requires a UriValue", "onto:testUriValue"),
    ]
    sorted_problems = result.problems
    assert isinstance(sorted_problems, SortedProblems)
    assert len(sorted_problems.unique_violations) == len(expected_info_tuples)
    assert not sorted_problems.user_warnings
    assert not sorted_problems.user_info
    assert not sorted_problems.unexpected_shacl_validation_components
    alphabetically_sorted = sorted(sorted_problems.unique_violations, key=lambda x: str(x.res_id))
    for one_result, expected_info in zip(alphabetically_sorted, expected_info_tuples):
        assert one_result.problem_type == ProblemType.VALUE_TYPE_MISMATCH
        assert one_result.res_id == expected_info[0]
        assert one_result.expected == expected_info[1]
        assert one_result.prop_name == expected_info[2]
    assert not _get_validation_status(sorted_problems, is_on_prod=True)
    assert not _get_validation_status(sorted_problems, is_on_prod=False)


@pytest.mark.usefixtures("create_generic_project")
def test_reformat_unique_value_violation(authentication) -> None:
    file = Path("testdata/validate-data/generic/unique_value_violation.xml")
    graphs, used_iris = prepare_data_for_validation_from_file(
        file, authentication, CONFIG.ignore_duplicate_files_warning
    )
    result = _validate_data(graphs, used_iris, CONFIG)
    assert not result.no_problems
    expected_ids = [
        "identical_values_LinkValue",
        "identical_values_listNode",
        "identical_values_richtext",
        "identical_values_valueAsString",
        "identical_values_valueHas",
    ]
    sorted_problems = result.problems
    assert isinstance(sorted_problems, SortedProblems)
    assert len(sorted_problems.unique_violations) == len(expected_ids)
    assert not sorted_problems.user_warnings
    assert not sorted_problems.user_info
    assert not sorted_problems.unexpected_shacl_validation_components
    alphabetically_sorted = sorted(sorted_problems.unique_violations, key=lambda x: str(x.res_id))
    for one_result, expected_id in zip(alphabetically_sorted, expected_ids):
        assert one_result.problem_type == ProblemType.DUPLICATE_VALUE
        assert one_result.res_id == expected_id
    assert not _get_validation_status(sorted_problems, is_on_prod=True)
    assert not _get_validation_status(sorted_problems, is_on_prod=False)


@pytest.mark.usefixtures("create_generic_project")
def test_reformat_file_value_violation(authentication) -> None:
    file = Path("testdata/validate-data/generic/file_value_violation.xml")
    graphs, used_iris = prepare_data_for_validation_from_file(
        file, authentication, CONFIG.ignore_duplicate_files_warning
    )
    result = _validate_data(graphs, used_iris, CONFIG)
    assert not result.no_problems
    expected_info_violation = [
        ("authorship_with_newline", ProblemType.GENERIC),
        ("copyright_holder_with_newline", ProblemType.GENERIC),
        ("empty_copyright_holder", ProblemType.INPUT_REGEX),
        ("empty_license", ProblemType.GENERIC),
        ("id_archive_missing", ProblemType.FILE_VALUE),
        ("id_archive_unknown", ProblemType.FILE_VALUE),
        ("id_audio_missing", ProblemType.FILE_VALUE),
        ("id_audio_unknown", ProblemType.FILE_VALUE),
        ("id_document_missing", ProblemType.FILE_VALUE),
        ("id_document_unknown", ProblemType.FILE_VALUE),
        ("id_resource_without_representation", ProblemType.FILE_VALUE_PROHIBITED),
        ("id_still_image_missing", ProblemType.FILE_VALUE),
        ("id_still_image_unknown", ProblemType.FILE_VALUE),
        ("id_text_missing", ProblemType.FILE_VALUE),
        ("id_text_unknown", ProblemType.FILE_VALUE),
        ("id_video_missing", ProblemType.FILE_VALUE),
        ("id_video_unknown", ProblemType.FILE_VALUE),
        ("id_wrong_file_type", ProblemType.FILE_VALUE),
        ("inexistent_license_iri", ProblemType.GENERIC),
        ("license_not_enabled", ProblemType.GENERIC),
        ("unknown_authorship_id", ProblemType.INPUT_REGEX),
    ]
    sorted_problems = result.problems
    assert isinstance(sorted_problems, SortedProblems)
    alphabetically_sorted_violations = sorted(sorted_problems.unique_violations, key=lambda x: str(x.res_id))
    assert len(sorted_problems.unique_violations) == len(expected_info_violation)
    assert not sorted_problems.user_warnings
    assert not sorted_problems.user_info
    assert not sorted_problems.unexpected_shacl_validation_components
    for one_result, expected_info in zip(alphabetically_sorted_violations, expected_info_violation):
        assert one_result.problem_type == expected_info[1]
        assert one_result.res_id == expected_info[0]
    assert not _get_validation_status(sorted_problems, is_on_prod=True)
    assert not _get_validation_status(sorted_problems, is_on_prod=False)


@pytest.mark.usefixtures("create_generic_project")
def test_reformat_dsp_inbuilt_violation(authentication) -> None:
    file = Path("testdata/validate-data/generic/dsp_inbuilt_violation.xml")
    graphs, used_iris = prepare_data_for_validation_from_file(
        file, authentication, CONFIG.ignore_duplicate_files_warning
    )
    result = _validate_data(graphs, used_iris, CONFIG)
    assert not result.no_problems
    expected_info_tuples = [
        ("audio_segment_target_is_video", ProblemType.LINK_TARGET_TYPE_MISMATCH),
        ("audio_segment_target_non_existent", ProblemType.INEXISTENT_LINKED_RESOURCE),
        ("link_obj_target_non_existent", ProblemType.INEXISTENT_LINKED_RESOURCE),
        ("missing_isPartOf", ProblemType.GENERIC),
        ("missing_seqnum", ProblemType.GENERIC),
        ("non_existent_permissions_bitstream", ProblemType.GENERIC),
        ("non_existent_permissions_resource", ProblemType.GENERIC),
        ("non_existent_permissions_value", ProblemType.GENERIC),
        ("region_invalid_geometry", ProblemType.INPUT_REGEX),
        ("region_isRegionOf_resource_does_not_exist", ProblemType.INEXISTENT_LINKED_RESOURCE),
        ("region_isRegionOf_resource_not_a_representation", ProblemType.LINK_TARGET_TYPE_MISMATCH),
        ("target_must_be_a_representation", ProblemType.LINK_TARGET_TYPE_MISMATCH),
        ("target_must_be_an_image_representation", ProblemType.LINK_TARGET_TYPE_MISMATCH),
        ("video_segment_start_larger_than_end", ProblemType.GENERIC),
        ("video_segment_target_is_audio", ProblemType.LINK_TARGET_TYPE_MISMATCH),
        ("video_segment_target_non_existent", ProblemType.INEXISTENT_LINKED_RESOURCE),
        ("video_segment_wrong_bounds", ProblemType.GENERIC),  # once for start that is less than zero
        ("video_segment_wrong_bounds", ProblemType.GENERIC),  # once for the end that is zero
    ]
    sorted_problems = result.problems
    assert isinstance(sorted_problems, SortedProblems)
    assert len(sorted_problems.unique_violations) == len(expected_info_tuples)
    assert not sorted_problems.user_info
    assert not sorted_problems.unexpected_shacl_validation_components
    alphabetically_sorted = sorted(sorted_problems.unique_violations, key=lambda x: str(x.res_id))
    for one_result, expected_info in zip(alphabetically_sorted, expected_info_tuples):
        assert one_result.problem_type == expected_info[1]
        assert one_result.res_id == expected_info[0]
    assert not _get_validation_status(sorted_problems, is_on_prod=True)
    assert not _get_validation_status(sorted_problems, is_on_prod=False)


if __name__ == "__main__":
    pytest.main([__file__])



import pandas as pd
import pytest
import regex

from dsp_tools.commands.excel2json.lists.utils import get_all_languages_for_columns
from dsp_tools.commands.excel2json.lists.utils import get_column_info
from dsp_tools.commands.excel2json.lists.utils import get_columns_of_preferred_lang
from dsp_tools.commands.excel2json.lists.utils import get_hierarchy_nums
from dsp_tools.commands.excel2json.lists.utils import get_lang_string_from_column_name
from dsp_tools.commands.excel2json.lists.utils import get_preferred_language_from_columns
from dsp_tools.error.exceptions import InputError


def test_get_lang_string_good() -> None:
    assert get_lang_string_from_column_name("en_1") == "en"


def test_get_lang_string_raises() -> None:
    assert not get_lang_string_from_column_name("ru_1")


class TestGetRemainingColumns:
    def test_with_matching_columns(self) -> None:
        columns = pd.Index(["en_1", "de_2", "fr_3"])
        assert get_hierarchy_nums(columns) == [1, 2, 3]

    def test_with_list_language_columns(self) -> None:
        columns = pd.Index(["en_1", "de_2", "fr_3", "de_list"])
        assert get_hierarchy_nums(columns) == [1, 2, 3]

    def test_with_non_numeric_columns(self) -> None:
        columns = pd.Index(["en_1", "de_2", "fr_3", "it_4", "rm_5", "en_other"])
        assert get_hierarchy_nums(columns) == [1, 2, 3, 4, 5]


def test_make_columns_no_comments() -> None:
    res = get_column_info(pd.Index(["en_list", "fr_1", "en_3", "de_3"]))
    assert res.preferred_lang == "en"
    assert set(res.list_cols) == {"en_list", "de_list", "fr_list"}
    assert not res.comment_cols
    assert len(res.node_cols) == 2
    assert set(res.node_cols[0].columns) == {"en_3", "de_3", "fr_3"}
    assert set(res.node_cols[1].columns) == {"en_1", "de_1", "fr_1"}


def test_make_columns_with_comments() -> None:
    res = get_column_info(pd.Index(["en_list", "fr_1", "en_3", "de_3", "fr_comments"]))
    assert res.preferred_lang == "en"
    assert set(res.list_cols) == {"en_list", "de_list", "fr_list"}
    assert set(res.comment_cols) == {"en_comments", "de_comments", "fr_comments"}
    assert len(res.node_cols) == 2
    assert set(res.node_cols[0].columns) == {"en_3", "de_3", "fr_3"}
    assert set(res.node_cols[1].columns) == {"en_1", "de_1", "fr_1"}


class TestGetAllLanguagesForColumns:
    def test_get_all_languages_for_columns_returns_correct_languages(self) -> None:
        columns = pd.Index(["en_1", "de_1", "fr_1", "it_1", "rm_1", "en_2"])
        ending = "1"
        assert get_all_languages_for_columns(columns, ending) == {"en", "de", "fr", "it", "rm"}

    def test_get_all_languages_for_columns_returns_empty_list_for_no_match(self) -> None:
        columns = pd.Index(["en_1", "de_1", "fr_1", "it_1", "rm_1", "en_2", "de_2", "fr_2", "it_2", "rm_2"])
        ending = "3"
        assert not get_all_languages_for_columns(columns, ending)


class TestGetPreferredLanguageForId:
    def test_get_preferred_language_for_id_en(self) -> None:
        columns = pd.Index(["en_1", "de_1", "fr_1", "it_1", "rm_1"])
        assert get_preferred_language_from_columns(columns, "1") == "en"

    def test_get_preferred_language_for_id_de(self) -> None:
        columns = pd.Index(["de_1", "fr_1", "it_1", "rm_1"])
        assert get_preferred_language_from_columns(columns, "1") == "de"

    def test_get_preferred_language_for_id_fr(self) -> None:
        columns = pd.Index(["fr_1", "it_1", "rm_1"])
        assert get_preferred_language_from_columns(columns, "1") == "fr"

    def test_get_preferred_language_for_id_it(self) -> None:
        columns = pd.Index(["it_1", "rm_1"])
        assert get_preferred_language_from_columns(columns, "1") == "it"

    def test_get_preferred_language_for_id_rm(self) -> None:
        columns = pd.Index(["rm_1"])
        assert get_preferred_language_from_columns(columns, "1") == "rm"

    def test_get_preferred_language_for_id_raises(self) -> None:
        columns = pd.Index(["es_1"])
        msg = regex.escape(
            "The columns may only contain the languages: 'en', 'de', 'fr', 'it', 'rm'.\nThe columns are: es_1"
        )
        with pytest.raises(InputError, match=msg):
            get_preferred_language_from_columns(columns, "1")


def test_get_columns_preferred_lang_returns_expected_columns() -> None:
    columns = pd.Index(["en_2", "de_1", "en_1", "it_1"])
    assert get_columns_of_preferred_lang(columns, "en") == ["en_1", "en_2"]


def test_get_columns_preferred_lang_returns_empty_list_for_no_match() -> None:
    columns = pd.Index(["de_1", "de_2", "it_1"])
    assert not get_columns_of_preferred_lang(columns, "en")



from dataclasses import dataclass
from typing import Any
from urllib.parse import quote_plus

import requests
from loguru import logger
from requests import RequestException

from dsp_tools.clients.authentication_client import AuthenticationClient
from dsp_tools.utils.request_utils import RequestParameters
from dsp_tools.utils.request_utils import log_request
from dsp_tools.utils.request_utils import log_response


@dataclass
class PermissionsClient:
    auth: AuthenticationClient
    proj_iri: str

    def get_project_doaps(self) -> list[dict[str, Any]]:
        params = RequestParameters(
            "GET",
            f"{self.auth.server}/admin/permissions/doap/{quote_plus(self.proj_iri)}",
            timeout=10,
            headers={"Accept": "application/json", "Authorization": f"Bearer {self.auth.get_token()}"},
        )
        log_request(params)
        try:
            response = requests.get(params.url, timeout=params.timeout, headers=params.headers)
            log_response(response)
        except RequestException:
            logger.exception("Error while retrieving existing DOAPs")
            return []
        res: list[dict[str, Any]] = response.json()["default_object_access_permissions"]
        return res

    def delete_doap(self, doap_iri: str) -> bool:
        params = RequestParameters(
            "DELETE",
            f"{self.auth.server}/admin/permissions/{quote_plus(doap_iri)}",
            timeout=10,
            headers={"Authorization": f"Bearer {self.auth.get_token()}"},
        )
        log_request(params)
        try:
            response = requests.delete(params.url, timeout=params.timeout, headers=params.headers)
            log_response(response)
        except RequestException:
            logger.exception("Error while deleting DOAP")
            return False
        return True

    def create_new_doap(self, payload: dict[str, Any]) -> bool:
        params = RequestParameters(
            "POST",
            f"{self.auth.server}/admin/permissions/doap",
            timeout=10,
            headers={"Authorization": f"Bearer {self.auth.get_token()}"},
            data=payload,
        )
        log_request(params)
        try:
            response = requests.post(params.url, timeout=params.timeout, headers=params.headers, json=params.data)
            log_response(response)
        except RequestException:
            logger.exception("Error while creating new DOAP")
            return False
        return True



import pytest
from lxml import etree

from dsp_tools.xmllib.internal.serialise_file_value import _serialise_metadata
from dsp_tools.xmllib.internal.serialise_file_value import serialise_file_value
from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.models.internal.file_values import FileValue
from dsp_tools.xmllib.models.internal.file_values import IIIFUri
from dsp_tools.xmllib.models.internal.file_values import Metadata
from dsp_tools.xmllib.models.licenses.recommended import LicenseRecommended


@pytest.fixture
def metadata_no_permissions() -> Metadata:
    return Metadata(
        license=LicenseRecommended.DSP.UNKNOWN,
        copyright_holder="copyright",
        authorship=("one", "one2"),
        permissions=Permissions.PROJECT_SPECIFIC_PERMISSIONS,
    )


def test_serialise_metadata(metadata_no_permissions: Metadata) -> None:
    expected = {
        "license": "http://rdfh.ch/licenses/unknown",
        "copyright-holder": "copyright",
        "authorship-id": "authorship_1",
    }
    result = _serialise_metadata(metadata_no_permissions, "authorship_1")
    assert result == expected


def test_serialise_metadata_with_permissions() -> None:
    meta = Metadata(
        license=LicenseRecommended.DSP.UNKNOWN,
        copyright_holder="copyright",
        authorship=tuple(["auth"]),
        permissions=Permissions.PUBLIC,
    )
    expected = {
        "license": "http://rdfh.ch/licenses/unknown",
        "copyright-holder": "copyright",
        "authorship-id": "authorship_2",
        "permissions": "public",
    }
    result = _serialise_metadata(meta, "authorship_2")
    assert result == expected


def test_serialise_file_value_bitstream(metadata_no_permissions: Metadata) -> None:
    val = FileValue("file.jpg", metadata_no_permissions, None)
    result = serialise_file_value(val, "authorship_1")
    expected = (
        b'<bitstream xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'license="http://rdfh.ch/licenses/unknown" '
        b'copyright-holder="copyright" '
        b'authorship-id="authorship_1"'
        b">file.jpg</bitstream>"
    )
    assert etree.tostring(result) == expected


def test_serialise_file_value_bitstream_with_comment(metadata_no_permissions: Metadata) -> None:
    val = FileValue("file.jpg", metadata_no_permissions, "comment")
    result = serialise_file_value(val, "authorship_1")
    expected = (
        b'<bitstream xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'license="http://rdfh.ch/licenses/unknown" '
        b'copyright-holder="copyright" '
        b'authorship-id="authorship_1" '
        b'comment="comment"'
        b">file.jpg</bitstream>"
    )
    assert etree.tostring(result) == expected


def test_serialise_file_value_iiif(metadata_no_permissions: Metadata) -> None:
    val = IIIFUri("https://example.org/image.jpg/full/1338%2C/0/default.jpg", metadata_no_permissions, None)
    result = serialise_file_value(val, "authorship_1")
    expected = (
        b'<iiif-uri xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'license="http://rdfh.ch/licenses/unknown" '
        b'copyright-holder="copyright" '
        b'authorship-id="authorship_1"'
        b">https://example.org/image.jpg/full/1338%2C/0/default.jpg</iiif-uri>"
    )
    assert etree.tostring(result) == expected


def test_serialise_file_value_iiif_with_comment(metadata_no_permissions: Metadata) -> None:
    val = IIIFUri("https://example.org/image.jpg/full/1338%2C/0/default.jpg", metadata_no_permissions, "comment")
    result = serialise_file_value(val, "authorship_1")
    expected = (
        b'<iiif-uri xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'license="http://rdfh.ch/licenses/unknown" '
        b'copyright-holder="copyright" '
        b'authorship-id="authorship_1" '
        b'comment="comment"'
        b">https://example.org/image.jpg/full/1338%2C/0/default.jpg</iiif-uri>"
    )
    assert etree.tostring(result) == expected


def test_serialise_file_value_iiif_without_legal_info() -> None:
    meta = Metadata(
        license=None,
        copyright_holder=None,
        authorship=None,
        permissions=Permissions.PROJECT_SPECIFIC_PERMISSIONS,
    )
    val = IIIFUri("https://example.org/image.jpg/full/1338%2C/0/default.jpg", meta, None)
    result = serialise_file_value(val, None)
    expected = (
        b'<iiif-uri xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"'
        b">https://example.org/image.jpg/full/1338%2C/0/default.jpg</iiif-uri>"
    )
    assert etree.tostring(result) == expected


if __name__ == "__main__":
    pytest.main([__file__])



# explicitly re-export imported stuff, so that API users can import it from this module
# (see https://mypy.readthedocs.io/en/stable/command_line.html#cmdoption-mypy-no-implicit-reexport)

from dsp_tools.commands.excel2xml.excel2xml_lib import *
from dsp_tools.commands.excel2xml.propertyelement import PropertyElement as PropertyElement
from dsp_tools.utils.data_formats.shared import check_notna as check_notna
from dsp_tools.utils.data_formats.shared import simplify_name as simplify_name



from __future__ import annotations

from html.entities import html5
from typing import Any

import regex

from dsp_tools.error.xmllib_warnings import MessageInfo
from dsp_tools.error.xmllib_warnings_util import emit_xmllib_input_info
from dsp_tools.error.xmllib_warnings_util import emit_xmllib_input_type_mismatch_warning
from dsp_tools.error.xmllib_warnings_util import raise_input_error
from dsp_tools.xmllib.internal.checkers import check_and_warn_if_a_string_contains_a_potentially_empty_value
from dsp_tools.xmllib.internal.checkers import is_nonempty_value_internal
from dsp_tools.xmllib.internal.constants import PREDEFINED_XML_ENTITIES


def check_and_get_corrected_comment(comment: Any, res_id: str | None, prop_name: str | None) -> str | None:
    """The input of comments may also be pd.NA or such. In our models we only want a string or None."""
    if is_nonempty_value_internal(comment):
        check_and_warn_if_a_string_contains_a_potentially_empty_value(
            value=comment,
            res_id=res_id,
            prop_name=prop_name,
            field="comment on value",
        )
        return str(comment)
    return None


def check_and_fix_is_non_empty_string(
    value: Any, res_id: str | None = None, prop_name: str | None = None, value_field: str | None = None
) -> str:
    """
    Emits warnings if the string is or looks empty.
    Converts a non-empty input into a string.
    Returns an empty string if the input is empty (eg. pd.NA, None, etc.).

    Args:
        value: input to check
        res_id: resource ID
        prop_name: property name
        value_field: field if it is not a property

    Returns:
        The value as string, if it is empty an empty string.
    """
    if is_nonempty_value_internal(value):
        check_and_warn_if_a_string_contains_a_potentially_empty_value(
            value=value,
            res_id=res_id,
            prop_name=prop_name,
            field=value_field,
        )
        return str(value)
    else:
        emit_xmllib_input_type_mismatch_warning(
            expected_type="non empty string",
            value=value,
            res_id=res_id,
            prop_name=prop_name,
            value_field=value_field,
        )
        return ""


def check_and_fix_collection_input(value: Any, prop_name: str, res_id: str) -> list[Any]:
    """
    To allow varied input but ensure consistent typing internally, collections are converted.
    If a collection is empty, a warning is emitted for the user.

    Args:
        value: Input value
        prop_name: Property name
        res_id: Resource ID

    Returns:
        The input as a list

    Raises:
        InputError: if the input is a dictionary
    """
    match value:
        case set() | list() | tuple():
            if len(value) == 0:
                msg_info = MessageInfo(
                    message="The input is empty. Please note that no values will be added to the resource.",
                    resource_id=res_id,
                    prop_name=prop_name,
                )
                emit_xmllib_input_info(msg_info)
            return list(value)
        case dict():
            msg_info = MessageInfo(
                message="The input is a dictionary. Only collections (list, set, tuple) are permissible.",
                resource_id=res_id,
                prop_name=prop_name,
            )
            raise_input_error(msg_info)
        case _:
            return [value]


def unescape_reserved_xml_chars(richtext: str) -> str:
    """
    This function unescapes characters that are reserved in an XML.

    Args:
        richtext: Text to be escaped

    Returns:
        Escaped string
    """
    richtext = regex.sub(r"&lt;", "<", richtext or "")
    richtext = regex.sub(r"&gt;", ">", richtext)
    richtext = regex.sub(r"&amp;", "&", richtext)
    return richtext


def numeric_entities(text: str) -> str:
    """
    Make a string suitable to be written to an XML file,
    by replacing all named HTML entities by their decimal numeric counterparts,
    except the ones that are predefined in the XML specification.
    Numeric HTML entities remain untouched.
    Unescaped special characters remain untouched.

    Args:
        text: original text

    Returns:
        text with all named entities replaced

    Examples:
        ```python
        result = xmllib.numeric_entities('&nbsp; &quot;')
        # result == '&#160; &quot;'
        ```
    """
    replacements: dict[str, str] = {}
    for match in regex.findall(r"&[0-9A-Za-z]+;", text):
        if match in PREDEFINED_XML_ENTITIES:
            continue
        char = html5[match[1:]]
        replacements[match] = f"&#{ord(char)};"
    text = regex.sub(
        r"&[0-9A-Za-z]+;",
        lambda x: replacements[x.group()] if x.group() not in PREDEFINED_XML_ENTITIES else x.group(),
        text,
    )
    return text



import json
from pathlib import Path
from typing import Any
from typing import Union

from dsp_tools.error.exceptions import BaseError


def parse_json_input(project_file_as_path_or_parsed: Union[str, Path, dict[str, Any]]) -> dict[str, Any]:
    """
    Check the input for a method that expects a JSON project definition, either as file path or as parsed JSON object:
    If it is parsed already, return it unchanged.
    If the input is a file path, parse it.

    Args:
        project_file_as_path_or_parsed: path to the JSON project definition, or parsed JSON object

    Raises:
        BaseError: if the input is invalid

    Returns:
        the parsed JSON object
    """
    project_definition: dict[str, Any] = {}
    if isinstance(project_file_as_path_or_parsed, dict):
        project_definition = project_file_as_path_or_parsed
    elif isinstance(project_file_as_path_or_parsed, str | Path) and Path(project_file_as_path_or_parsed).exists():
        with open(project_file_as_path_or_parsed, encoding="utf-8") as f:
            try:
                project_definition = json.load(f)
            except json.JSONDecodeError as e:
                msg = f"The input file '{project_file_as_path_or_parsed}' cannot be parsed to a JSON object."
                raise BaseError(msg) from e
    else:
        raise BaseError("Invalid input: The input must be a path to a JSON file or a parsed JSON object.")
    return project_definition



import pytest

from dsp_tools.commands.validate_data.models.api_responses import OneList
from dsp_tools.commands.validate_data.models.api_responses import OneNode


@pytest.fixture
def one_list() -> OneList:
    return OneList(
        list_iri="http://rdfh.ch/lists/9999/list1",
        list_name="firstList",
        nodes=[
            OneNode("n1", "http://rdfh.ch/lists/9999/n1"),
            OneNode("n1.1", "http://rdfh.ch/lists/9999/n11"),
            OneNode("n1.1.1", "http://rdfh.ch/lists/9999/n111"),
        ],
    )


class TestOneList:
    def test_hlist(self, one_list: OneList) -> None:
        result = one_list.hlist()
        expected = '"hlist=<http://rdfh.ch/lists/9999/list1>"'
        assert result == expected



import json
from typing import Any
from typing import cast

import pytest

from dsp_tools.commands.excel2json.project import _old_create_project_json
from dsp_tools.commands.excel2json.project import _old_validate_folder_structure_and_get_filenames


@pytest.fixture
def expected_project() -> dict[str, Any]:
    with open("testdata/excel2json/excel2json-expected-output.json", encoding="utf-8") as f:
        proj = json.load(f)
        return cast(dict[str, Any], proj)


@pytest.fixture
def returned_project() -> dict[str, Any]:
    excel_folder = "testdata/excel2json/old_excel2json_files"
    listfolder, onto_folders = _old_validate_folder_structure_and_get_filenames(excel_folder)
    _, project = _old_create_project_json(excel_folder, listfolder, onto_folders)
    return project


def test_same_keys(returned_project: dict[str, Any], expected_project: dict[str, Any]) -> None:
    assert set(returned_project["project"]) == set(expected_project["project"])


def test_lists(returned_project: dict[str, Any], expected_project: dict[str, Any]) -> None:
    res_lists = sorted(returned_project["project"]["lists"], key=lambda x: x["name"])
    expected_lists = sorted(expected_project["project"]["lists"], key=lambda x: x["name"])
    for r, e in zip(res_lists, expected_lists):
        assert r == e, r["name"]


def test_ontology_number(returned_project: dict[str, Any], expected_project: dict[str, Any]) -> None:
    res_ontos = returned_project["project"]["ontologies"]
    expected_onto = expected_project["project"]["ontologies"]
    assert len(res_ontos) == 1
    assert len(expected_onto) == 1


def test_properties(returned_project: dict[str, Any], expected_project: dict[str, Any]) -> None:
    res_props = sorted(returned_project["project"]["ontologies"][0]["properties"], key=lambda x: x["name"])
    ex_props = sorted(expected_project["project"]["ontologies"][0]["properties"], key=lambda x: x["name"])
    for r, e in zip(res_props, ex_props):
        assert r == e, r["name"]


def test_resources(returned_project: dict[str, Any], expected_project: dict[str, Any]) -> None:
    res_resources = sorted(returned_project["project"]["ontologies"][0]["resources"], key=lambda x: x["name"])
    ex_resources = sorted(expected_project["project"]["ontologies"][0]["resources"], key=lambda x: x["name"])
    for r, e in zip(res_resources, ex_resources):
        assert r == e, r["name"]


def test_entire_project(returned_project: dict[str, Any], expected_project: dict[str, Any]) -> None:
    assert returned_project == expected_project


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from dsp_tools.xmllib.models.licenses.recommended import License


class LicenseOther:
    """
    Pre-defined licenses that are available in DSP.
    If the user is free to choose a license, we recommend to choose from our [recommended licenses](https://docs.dasch.swiss/latest/DSP-TOOLS/xmllib-api-reference/licenses/recommended/).

    - `Public`: [See `Public` for details](https://docs.dasch.swiss/latest/DSP-TOOLS/xmllib-api-reference/licenses/other/#xmllib.models.licenses.other.Public).
    - `Various`: [See `Various` for details](https://docs.dasch.swiss/latest/DSP-TOOLS/xmllib-api-reference/licenses/other/#xmllib.models.licenses.other.Various).

    Examples:
        ```python
        # adding a BORIS Standard License to a file
        resource = resource.add_file(
            filename="images/cat.jpg",
            license=xmllib.LicenseOther.Various.BORIS_STANDARD,
            copyright_holder="Meow University",
            authorship=["Kitty Meow"],
        )
        ```
    """

    Public: type[Public]
    Various: type[Various]


class Public(License):
    """
    Pre-defined public domain licenses.
    [See the API documentation for details about the licenses.](https://docs.dasch.swiss/latest/DSP-API/01-introduction/legal-info/#license)

    - `CC_0_1_0`: CC0 1.0 Universal
    - `CC_PDM_1_0`: Public Domain Mark 1.0 Universal
    """

    CC_0_1_0 = "http://rdfh.ch/licenses/cc-0-1.0"
    CC_PDM_1_0 = "http://rdfh.ch/licenses/cc-pdm-1.0"


class Various(License):
    """
    A collection of various, pre-defined licenses.
    [See the API documentation for details about the licenses.](https://docs.dasch.swiss/latest/DSP-API/01-introduction/legal-info/#license)

    - `BORIS_STANDARD`: BORIS Standard License
    - `FRANCE_OUVERTE`: LICENCE OUVERTE 2.0
    """

    BORIS_STANDARD = "http://rdfh.ch/licenses/boris"
    FRANCE_OUVERTE = "http://rdfh.ch/licenses/open-licence-2.0"


LicenseOther.Public = Public
LicenseOther.Various = Various



from loguru import logger

from dsp_tools.commands.project.models.permissions_client import PermissionsClient

USER_IRI_PREFIX = "http://www.knora.org/ontology/knora-admin#"


def create_default_permissions(perm_client: PermissionsClient, default_permissions: str) -> bool:
    logger.info("Set default permissions...")
    print("Set default permissions...")
    if not _delete_existing_doaps(perm_client):
        print("WARNING: Cannot delete the existing default permissions")
        logger.warning("Cannot delete the existing default permissions")
        return False
    if not _create_new_doap(perm_client, default_permissions):
        print("WARNING: Cannot create default permissions")
        logger.warning("Cannot create default permissions")
        return False
    logger.info("Default permissions have been set")
    print("Default permissions have been set")
    return True


def _delete_existing_doaps(perm_client: PermissionsClient) -> bool:
    if not (doaps := perm_client.get_project_doaps()):
        return False
    existing_doap_iris: list[str] = [x["iri"] for x in doaps]
    for iri in existing_doap_iris:
        if not perm_client.delete_doap(iri):
            # don't continue with the others, it's better to stop DOAP handling immediately, to avoid a mess
            return False
    return True


def _create_new_doap(perm_client: PermissionsClient, default_permissions: str) -> bool:
    perm = [
        {"additionalInformation": f"{USER_IRI_PREFIX}ProjectAdmin", "name": "CR", "permissionCode": None},
        {"additionalInformation": f"{USER_IRI_PREFIX}ProjectMember", "name": "D", "permissionCode": None},
    ]
    if default_permissions == "public":
        perm.append({"additionalInformation": f"{USER_IRI_PREFIX}KnownUser", "name": "V", "permissionCode": None})
        perm.append({"additionalInformation": f"{USER_IRI_PREFIX}UnknownUser", "name": "V", "permissionCode": None})
    payload = {
        "forGroup": f"{USER_IRI_PREFIX}ProjectMember",
        "forProject": perm_client.proj_iri,
        "hasPermissions": perm,
    }
    return perm_client.create_new_doap(payload)



from pathlib import Path

import pandas as pd

from dsp_tools.commands.ingest_xmlupload.upload_files.upload_failures import UploadFailure
from dsp_tools.commands.ingest_xmlupload.upload_files.upload_failures import UploadFailures

SHORTCODE = "0001"
DSP_INGEST_URL = "https://ingest.test.dasch.swiss"


def test_one_failure() -> None:
    num_of_all_files = 3
    failures = [UploadFailure(Path("path1"), "reason1")]
    aggregated_failures = UploadFailures(failures, num_of_all_files, SHORTCODE, DSP_INGEST_URL)
    msg = aggregated_failures.execute_error_protocol()
    expected = (
        f"Uploaded 2/3 files onto server {DSP_INGEST_URL}. Failed to upload the following 1 files:\n- path1: reason1"
    )
    assert msg == expected


def test_several_failures() -> None:
    num_of_all_files = 5
    failures = [
        UploadFailure(Path("path1"), "reason1"),
        UploadFailure(Path("path2"), "reason2", 404, "not found"),
        UploadFailure(Path("path3"), "reason3"),
    ]
    aggregated_failures = UploadFailures(failures, num_of_all_files, SHORTCODE, DSP_INGEST_URL)
    msg = aggregated_failures.execute_error_protocol()
    expected = (
        f"Uploaded 2/5 files onto server {DSP_INGEST_URL}. Failed to upload the following 3 files:\n"
        "- path1: reason1\n"
        "- path2: reason2\n"
        "    - Status code: 404\n"
        "    - Response text: not found\n"
        "- path3: reason3"
    )
    assert msg == expected


def test_all_failures() -> None:
    num_of_all_files = 3
    failures = [
        UploadFailure(Path("path1"), "reason1"),
        UploadFailure(Path("path2"), "reason2"),
        UploadFailure(Path("path3"), "reason3"),
    ]
    aggregated_failures = UploadFailures(failures, num_of_all_files, SHORTCODE, DSP_INGEST_URL)
    msg = aggregated_failures.execute_error_protocol()
    expected = (
        f"Uploaded 0/3 files onto server {DSP_INGEST_URL}. Failed to upload the following 3 files:\n"
        "- path1: reason1\n"
        "- path2: reason2\n"
        "- path3: reason3"
    )
    assert msg == expected


def test_several_failures_with_file_writing() -> None:
    num_of_all_files = 5
    failures = [
        UploadFailure(Path("path1"), "reason1"),
        UploadFailure(Path("path2"), "reason2"),
        UploadFailure(Path("path3"), "reason3"),
    ]
    aggregated_failures = UploadFailures(failures, num_of_all_files, SHORTCODE, DSP_INGEST_URL, maximum_prints=2)
    msg = aggregated_failures.execute_error_protocol()
    expected_output_path = Path(f"upload_failures_{SHORTCODE}_ingest.test.dasch.swiss.csv")
    expected = (
        f"Uploaded 2/5 files onto server {DSP_INGEST_URL}. Failed to upload 3 files. "
        f"The full list of failed files has been saved to '{expected_output_path}'."
    )
    assert msg == expected

    df = pd.read_csv(expected_output_path)
    assert df["Filepath"].tolist() == ["path1", "path2", "path3"]
    assert df["Reason"].tolist() == ["reason1", "reason2", "reason3"]
    expected_output_path.unlink()



from collections import defaultdict

from dsp_tools.cli.args import ValidateDataConfig
from dsp_tools.commands.validate_data.constants import MAXIMUM_DUPLICATE_FILE_PATHS
from dsp_tools.commands.validate_data.models.validation import DuplicateFileResult
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedResource


def check_for_duplicate_files(
    parsed_resources: list[ParsedResource], config: ValidateDataConfig
) -> DuplicateFileResult:
    """
    Too many duplicate filepaths in the data may cause the SHACL validator to crash.
    If one file is referenced n times, this produces n * (n-1) validation errors.
    Therefore, this programmatic pre-validation prevents crashes during later validation steps.

    Args:
        parsed_resources: Resources to check
        config: Validate data config

    Returns:
        Results for the user and decisions how the program should continue
    """
    count_dict = _get_filepaths_with_more_than_one_usage(parsed_resources)
    max_count_of_duplicates_reached = _determine_if_max_count_has_been_reached(count_dict)
    return _determine_duplicate_file_result(count_dict, max_count_of_duplicates_reached, config.is_on_prod_server)


def _get_filepaths_with_more_than_one_usage(parsed_resources: list[ParsedResource]) -> dict[str, int]:
    count_dict: dict[str, int] = defaultdict(int)
    for res in parsed_resources:
        if res.file_value and res.file_value.value:
            count_dict[res.file_value.value] += 1
    return {f_path: count for f_path, count in count_dict.items() if count > 1}


def _determine_if_max_count_has_been_reached(path_count: dict[str, int]) -> bool:
    return any([cnt > MAXIMUM_DUPLICATE_FILE_PATHS for cnt in path_count.values()])


def _determine_duplicate_file_result(
    count_dict: dict[str, int], max_count_of_duplicates_reached: bool, is_on_prod_server: bool
) -> DuplicateFileResult:
    if not max_count_of_duplicates_reached:
        return DuplicateFileResult(
            user_msg=None,
            duplicate_files_must_be_ignored=False,
            should_continue=True,
        )
    msg = _get_duplicate_msg(count_dict, is_on_prod_server)
    should_continue = not is_on_prod_server
    return DuplicateFileResult(
        user_msg=msg,
        duplicate_files_must_be_ignored=True,
        should_continue=should_continue,
    )


def _get_duplicate_msg(count_dict: dict[str, int], is_on_prod: bool) -> str:
    file_paths = sorted([f"{count} - '{path}'" for path, count in count_dict.items()])
    msg = (
        f"{len(count_dict)} file(s) were used multiple times in your data. "
        f"Due to the large number of duplicates they cannot be included in the schema validation.\n"
    )
    if is_on_prod:
        msg += (
            "Since you are on a production server, the validation or xmlupload cannot continue. "
            "If you wish to upload duplicate images, use the designated flag to ignore them."
        )
    else:
        msg += (
            "Since you are on a test environment, "
            "the validation or xmlupload will continue without the duplicate check. "
            "Please note that this is not allowed on a production server."
        )
    msg += (
        f"\nThe following filepaths are used more than once, "
        f"result displayed as: file count - 'file path'\n{' | '.join(file_paths)}"
    )
    return msg



from lxml import etree

from dsp_tools.error.exceptions import BaseError
from dsp_tools.xmllib.internal.constants import DASCH_SCHEMA
from dsp_tools.xmllib.internal.constants import XML_NAMESPACE_MAP
from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.models.internal.file_values import AbstractFileValue
from dsp_tools.xmllib.models.internal.file_values import FileValue
from dsp_tools.xmllib.models.internal.file_values import IIIFUri
from dsp_tools.xmllib.models.internal.file_values import Metadata
from dsp_tools.xmllib.value_checkers import is_nonempty_value


def serialise_file_value(file_value: AbstractFileValue, authorship_id: str | None) -> etree._Element:
    """
    Serialise the file value of a resource.

    Args:
        file_value: File Value
        authorship_id: ID referencing the authorship

    Returns:
        File Value as etree

    Raises:
        BaseError: in case of an unknown class
    """
    if isinstance(file_value, FileValue):
        return _serialise_file_value(file_value, authorship_id, "bitstream")
    elif isinstance(file_value, IIIFUri):
        return _serialise_file_value(file_value, authorship_id, "iiif-uri")
    else:
        raise BaseError(
            f"file_value must be either a FileValue or a IIIFUri, but you provided {file_value.__class__.__name__}"
        )


def _serialise_file_value(value: AbstractFileValue, authorship_id: str | None, tag_name: str) -> etree._Element:
    attribs = _serialise_metadata(value.metadata, authorship_id)
    if is_nonempty_value(value.comment):
        attribs["comment"] = str(value.comment)
    ele = etree.Element(f"{DASCH_SCHEMA}{tag_name}", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
    ele.text = str(value.value)
    return ele


def _serialise_metadata(metadata: Metadata, authorship_id: str | None) -> dict[str, str]:
    attribs = {}
    if metadata.license:
        attribs["license"] = str(metadata.license)
    if metadata.copyright_holder:
        attribs["copyright-holder"] = metadata.copyright_holder
    if authorship_id:
        attribs["authorship-id"] = authorship_id
    if metadata.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
        attribs["permissions"] = metadata.permissions.value
    return attribs



# mypy: disable-error-code="method-assign,no-untyped-def"

import pytest

from dsp_tools.commands.validate_data.models.api_responses import OneList
from dsp_tools.commands.validate_data.models.api_responses import OneNode
from dsp_tools.commands.validate_data.models.input_problems import InputProblem
from dsp_tools.commands.validate_data.models.input_problems import ProblemType
from dsp_tools.commands.validate_data.models.input_problems import Severity
from dsp_tools.commands.validate_data.models.input_problems import SortedProblems
from dsp_tools.commands.validate_data.prepare_data.prepare_data import _make_list_lookup
from dsp_tools.commands.validate_data.validate_data import _get_validation_status


def test_make_list_lookup():
    all_lists = [
        OneList("IRI1", "list1", [OneNode("l1n1", "IRIl1n1"), OneNode("l1n2", "IRIl1n2")]),
        OneList("IRI2", "list2", [OneNode("l2n1", "IRIl2n1")]),
    ]
    result = _make_list_lookup(all_lists)
    expected_lookup = {
        ("", "IRIl1n1"): "IRIl1n1",
        ("", "IRIl1n2"): "IRIl1n2",
        ("", "IRIl2n1"): "IRIl2n1",
        ("list1", "l1n1"): "IRIl1n1",
        ("list1", "l1n2"): "IRIl1n2",
        ("list2", "l2n1"): "IRIl2n1",
    }
    assert result.lists == expected_lookup


class TestGetValidationStatus:
    def test_not_prod_problems(self):
        problems = SortedProblems([InputProblem(ProblemType.GENERIC, "", "", "", Severity.VIOLATION)], [], [], [])
        validation_passed = _get_validation_status(problems, is_on_prod=False)
        assert not validation_passed

    def test_not_prod_problems_and_unexpected(self):
        problems = SortedProblems([InputProblem(ProblemType.GENERIC, "", "", "", Severity.VIOLATION)], [], [], ["unex"])
        validation_passed = _get_validation_status(problems, is_on_prod=False)
        assert not validation_passed

    def test_not_prod_warnings(self):
        problems = SortedProblems([], [InputProblem(ProblemType.GENERIC, "", "", "", Severity.VIOLATION)], [], [])
        validation_passed = _get_validation_status(problems, is_on_prod=False)
        assert validation_passed

    def test_not_prod_info(self):
        problems = SortedProblems([], [], [InputProblem(ProblemType.GENERIC, "", "", "", Severity.VIOLATION)], [])
        validation_passed = _get_validation_status(problems, is_on_prod=False)
        assert validation_passed

    def test_not_prod_unexpected(self):
        problems = SortedProblems([], [], [], ["unexpected"])
        validation_passed = _get_validation_status(problems, is_on_prod=False)
        assert not validation_passed


if __name__ == "__main__":
    pytest.main([__file__])



from uuid import uuid4

import pytest
import regex
from rdflib import RDF
from rdflib import XSD
from rdflib import BNode
from rdflib import Literal
from rdflib import Namespace
from rdflib import URIRef

from dsp_tools.commands.xmlupload.iri_resolver import IriResolver
from dsp_tools.commands.xmlupload.make_rdf_graph.make_values import _make_one_value_graph
from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue
from dsp_tools.commands.xmlupload.models.lookup_models import IRILookups
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.models.permission import PermissionValue
from dsp_tools.commands.xmlupload.models.processed.values import IntervalFloats
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedBoolean
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedColor
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedDate
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedDecimal
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedGeometry
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedGeoname
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedInt
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedInterval
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedLink
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedList
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedRichtext
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedSimpleText
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedTime
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedUri
from dsp_tools.error.exceptions import BaseError
from dsp_tools.utils.data_formats.date_util import Calendar
from dsp_tools.utils.data_formats.date_util import Date
from dsp_tools.utils.data_formats.date_util import Era
from dsp_tools.utils.data_formats.date_util import SingleDate
from dsp_tools.utils.rdflib_constants import KNORA_API

ONTO = Namespace("http://0.0.0.0:3333/ontology/9999/onto/v2#")

PERMISSION_LITERAL = Literal("CR knora-admin:ProjectAdmin", datatype=XSD.string)
RES_ONE_URI = URIRef("http://rdfh.ch/9999/res_one")


DUMMY_PERMISSION = Permissions({PermissionValue.CR: ["knora-admin:ProjectAdmin"]})


def absolute_iri(prop: str) -> str:
    return f"http://0.0.0.0:3333/ontology/9999/onto/v2#{prop}"


@pytest.fixture
def lookups() -> IRILookups:
    return IRILookups(
        project_iri=URIRef("http://rdfh.ch/9999/project"),
        id_to_iri=IriResolver({"res_one": "http://rdfh.ch/9999/res_one"}),
    )


class TestMakeOneValueGraphSuccess:
    def test_boolean(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = ProcessedBoolean(True, absolute_iri("isTrueOrFalse"), None, DUMMY_PERMISSION)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 4
        val_bn = next(result.objects(res_bn, ONTO.isTrueOrFalse))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.BooleanValue
        value = next(result.objects(val_bn, KNORA_API.booleanValueAsBoolean))
        assert value == Literal(True, datatype=XSD.boolean)
        permissions = next(result.objects(val_bn, KNORA_API.hasPermissions))
        assert permissions == PERMISSION_LITERAL

    def test_color(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = ProcessedColor("#5d1f1e", absolute_iri("hasColor"), None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, ONTO.hasColor))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.ColorValue
        value = next(result.objects(val_bn, KNORA_API.colorValueAsColor))
        assert value == Literal("#5d1f1e", datatype=XSD.string)

    def test_decimal(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = ProcessedDecimal(2.718281828459, absolute_iri("hasDecimal"), "Eulersche Zahl", None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 4
        val_bn = next(result.objects(res_bn, ONTO.hasDecimal))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.DecimalValue
        value = next(result.objects(val_bn, KNORA_API.decimalValueAsDecimal))
        assert value == Literal("2.718281828459", datatype=XSD.decimal)
        comment = next(result.objects(val_bn, KNORA_API.valueHasComment))
        assert comment == Literal("Eulersche Zahl", datatype=XSD.string)

    def test_geometry(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = ProcessedGeometry(
            '{"status": "active", "type": "polygon", "lineWidth": 5, '
            '"points": [{"x": 0.4, "y": 0.6}, {"x": 0.5, "y": 0.9}, {"x": 0.8, "y": 0.9}, {"x": 0.7, "y": 0.6}]}',
            absolute_iri("hasGeometry"),
            None,
            None,
        )
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, ONTO.hasGeometry))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.GeomValue
        value = next(result.objects(val_bn, KNORA_API.geometryValueAsGeometry))
        assert isinstance(value, Literal)

    def test_geoname(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = ProcessedGeoname("5416656", absolute_iri("hasGeoname"), None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, ONTO.hasGeoname))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.GeonameValue
        value = next(result.objects(val_bn, KNORA_API.geonameValueAsGeonameCode))
        assert value == Literal("5416656", datatype=XSD.string)

    def test_integer(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = ProcessedInt(1, absolute_iri("hasInteger"), "comment", None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 4
        val = next(result.subjects(KNORA_API.intValueAsInt, Literal("1", datatype=XSD.integer)))
        assert next(result.objects(val, RDF.type)) == KNORA_API.IntValue
        assert next(result.subjects(ONTO.hasInteger, val)) == res_bn
        comment = next(result.objects(val, KNORA_API.valueHasComment))
        assert comment == Literal("comment", datatype=XSD.string)

    def test_time(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = ProcessedTime("2019-10-23T13:45:12.01-14:00", absolute_iri("hasTime"), None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, ONTO.hasTime))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.TimeValue
        value = next(result.objects(val_bn, KNORA_API.timeValueAsTimeStamp))
        assert value == Literal("2019-10-23T13:45:12.01-14:00", datatype=XSD.dateTimeStamp)

    def test_uri(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = ProcessedUri("https://dasch.swiss", absolute_iri("hasUri"), None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, ONTO.hasUri))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.UriValue
        value = next(result.objects(val_bn, KNORA_API.uriValueAsUri))
        assert value == Literal("https://dasch.swiss", datatype=XSD.anyURI)

    def test_list(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = ProcessedList("http://rdfh.ch/9999/node", absolute_iri("hasListItem"), None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, ONTO.hasListItem))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.ListValue
        value = next(result.objects(val_bn, KNORA_API.listValueAsListNode))
        assert value == URIRef("http://rdfh.ch/9999/node")

    def test_resptr(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = ProcessedLink("res_one", absolute_iri("hasResource"), None, None, str(uuid4()))
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, ONTO.hasResourceValue))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.LinkValue
        value = next(result.objects(val_bn, KNORA_API.linkValueHasTargetIri))
        assert value == RES_ONE_URI

    def test_simpletext(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = ProcessedSimpleText("Text", absolute_iri("hasSimpleText"), None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, ONTO.hasSimpleText))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.TextValue
        value = next(result.objects(val_bn, KNORA_API.valueAsString))
        assert value == Literal("Text", datatype=XSD.string)

    def test_richtext(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = ProcessedRichtext(
            FormattedTextValue("Text"),
            absolute_iri("hasRichtext"),
            None,
            DUMMY_PERMISSION,
            resource_references=set(),
            value_uuid=str(uuid4()),
        )
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 5
        val_bn = next(result.objects(res_bn, ONTO.hasRichtext))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.TextValue
        value = next(result.objects(val_bn, KNORA_API.textValueAsXml))
        # FormattedTextValue adds a newline after the xml declaration
        assert value == Literal('<?xml version="1.0" encoding="UTF-8"?>\n<text>Text</text>', datatype=XSD.string)
        mapping = next(result.objects(val_bn, KNORA_API.textValueHasMapping))
        assert mapping == URIRef("http://rdfh.ch/standoff/mappings/StandardMapping")
        permissions = next(result.objects(val_bn, KNORA_API.hasPermissions))
        assert permissions == PERMISSION_LITERAL

    def test_richtext_with_reference(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        text = 'Comment with <a class="salsah-link" href="IRI:res_one:IRI">link to res_one'
        prop = ProcessedRichtext(
            FormattedTextValue(text),
            absolute_iri("hasRichtext"),
            None,
            None,
            resource_references=set("res_one"),
            value_uuid=str(uuid4()),
        )
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 4
        val_bn = next(result.objects(res_bn, ONTO.hasRichtext))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.TextValue
        value = next(result.objects(val_bn, KNORA_API.textValueAsXml))
        expected_text = (
            '<?xml version="1.0" encoding="UTF-8"?>\n'  # FormattedTextValue adds a newline after the xml declaration
            '<text>Comment with <a class="salsah-link" href="http://rdfh.ch/9999/res_one">'
            "link to res_one</text>"
        )
        assert value == Literal(expected_text, datatype=XSD.string)
        mapping = next(result.objects(val_bn, KNORA_API.textValueHasMapping))
        assert mapping == URIRef("http://rdfh.ch/standoff/mappings/StandardMapping")

    def test_date(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        date = Date(
            calendar=Calendar.GREGORIAN,
            start=SingleDate(era=Era.AD, year=476, month=9, day=4),
            end=SingleDate(era=Era.AD, year=477, month=None, day=None),
        )
        prop = ProcessedDate(date, absolute_iri("hasDate"), None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 9
        val_bn = next(result.objects(res_bn, ONTO.hasDate))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.DateValue
        calendar = next(result.objects(val_bn, KNORA_API.dateValueHasCalendar))
        assert calendar == Literal("GREGORIAN", datatype=XSD.string)
        start_day = next(result.objects(val_bn, KNORA_API.dateValueHasStartDay))
        assert start_day == Literal("4", datatype=XSD.integer)
        start_month = next(result.objects(val_bn, KNORA_API.dateValueHasStartMonth))
        assert start_month == Literal("9", datatype=XSD.integer)
        start_year = next(result.objects(val_bn, KNORA_API.dateValueHasStartYear))
        assert start_year == Literal("476", datatype=XSD.integer)
        start_era = next(result.objects(val_bn, KNORA_API.dateValueHasStartEra))
        assert start_era == Literal("AD", datatype=XSD.string)
        end_year = next(result.objects(val_bn, KNORA_API.dateValueHasEndYear))
        assert end_year == Literal("477", datatype=XSD.integer)
        start_era = next(result.objects(val_bn, KNORA_API.dateValueHasEndEra))
        assert start_era == Literal("AD", datatype=XSD.string)

    def test_interval(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        interval = IntervalFloats(0.1, 0.234)
        prop = ProcessedInterval(interval, "http://api.knora.org/ontology/knora-api/v2#hasSegmentBounds", None, None)
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 4
        val_bn = next(result.objects(res_bn, KNORA_API.hasSegmentBounds))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.IntervalValue
        start = next(result.objects(val_bn, KNORA_API.intervalValueHasStart))
        assert start == Literal("0.1", datatype=XSD.decimal)
        end = next(result.objects(val_bn, KNORA_API.intervalValueHasEnd))
        assert end == Literal("0.234", datatype=XSD.decimal)

    def test_segment_of_video(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = ProcessedLink(
            "res_one",
            "http://api.knora.org/ontology/knora-api/v2#isVideoSegmentOf",
            None,
            None,
            value_uuid=str(uuid4()),
        )
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, KNORA_API.isVideoSegmentOfValue))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.LinkValue
        value = next(result.objects(val_bn, KNORA_API.linkValueHasTargetIri))
        assert value == RES_ONE_URI

    def test_segment_of_audio(self, lookups: IRILookups) -> None:
        res_bn = BNode()
        prop = ProcessedLink(
            "res_one",
            "http://api.knora.org/ontology/knora-api/v2#isAudioSegmentOf",
            None,
            None,
            value_uuid=str(uuid4()),
        )
        result = _make_one_value_graph(prop, res_bn, lookups)
        assert len(result) == 3
        val_bn = next(result.objects(res_bn, KNORA_API.isAudioSegmentOfValue))
        rdf_type = next(result.objects(val_bn, RDF.type))
        assert rdf_type == KNORA_API.LinkValue
        value = next(result.objects(val_bn, KNORA_API.linkValueHasTargetIri))
        assert value == RES_ONE_URI


def test_link_target_not_found(lookups: IRILookups) -> None:
    res_bn = BNode()
    prop = ProcessedLink("non_existing", absolute_iri("hasResource"), None, None, value_uuid=str(uuid4()))
    err_str = regex.escape(
        "Could not find the ID non_existing in the id2iri mapping. "
        "This is probably because the resource 'non_existing' could not be created. "
        "See warnings.log for more information."
    )
    with pytest.raises(BaseError, match=err_str):
        _make_one_value_graph(prop, res_bn, lookups)


def test_richtext_with_reference_not_found(lookups: IRILookups) -> None:
    res_bn = BNode()
    text = 'Comment with <a class="salsah-link" href="IRI:nonExisingReference:IRI">link to res_one'
    prop = ProcessedRichtext(
        FormattedTextValue(text),
        absolute_iri("hasRichtext"),
        None,
        None,
        resource_references=set("nonExisingReference"),
        value_uuid=str(uuid4()),
    )
    err_str = regex.escape("Internal ID 'nonExisingReference' could not be resolved to an IRI")
    with pytest.raises(BaseError, match=err_str):
        _make_one_value_graph(prop, res_bn, lookups)



from typing import Any
from typing import Optional

import regex
from loguru import logger

from dsp_tools.clients.connection import Connection
from dsp_tools.commands.project.legacy_models.context import Context
from dsp_tools.commands.project.legacy_models.helpers import Cardinality
from dsp_tools.commands.project.legacy_models.ontology import Ontology
from dsp_tools.commands.project.legacy_models.project import Project
from dsp_tools.commands.project.legacy_models.propertyclass import PropertyClass
from dsp_tools.commands.project.legacy_models.resourceclass import ResourceClass
from dsp_tools.error.exceptions import BaseError
from dsp_tools.error.exceptions import InputError
from dsp_tools.legacy_models.datetimestamp import DateTimeStamp
from dsp_tools.legacy_models.langstring import LangString


def create_ontologies(
    con: Connection,
    context: Context,
    knora_api_prefix: str,
    names_and_iris_of_list_nodes: dict[str, Any],
    ontology_definitions: list[dict[str, Any]],
    project_remote: Project,
    verbose: bool,
) -> bool:
    """
    Iterates over the ontologies in a JSON project file and creates the ontologies that don't exist on the DSP server
    yet. For every ontology, it first creates the resource classes, then the properties, and then adds the cardinalities
    to the resource classes.

    Args:
        con: Connection to the DSP server
        context: prefixes and the ontology IRIs they stand for
        knora_api_prefix: the prefix that stands for the knora-api ontology
        names_and_iris_of_list_nodes: IRIs of list nodes that were already created and are available on the DSP server
        ontology_definitions: the "ontologies" section of the parsed JSON project file
        project_remote: representation of the project on the DSP server
        verbose: verbose switch

    Raises:
        InputError: if an error occurs during the creation of an ontology.
        All other errors are printed, the process continues, but the success status will be false.

    Returns:
        True if everything went smoothly, False otherwise
    """

    overall_success = True

    print("Create ontologies...")
    logger.info("Create ontologies...")
    try:
        project_ontologies = Ontology.getProjectOntologies(con=con, project_id=str(project_remote.iri))
    except BaseError:
        err_msg = "Unable to retrieve remote ontologies. Cannot check if your ontology already exists."
        print("WARNING: {err_msg}")
        logger.exception(err_msg)
        project_ontologies = []

    for ontology_definition in ontology_definitions:
        ontology_remote = _create_ontology(
            onto_name=ontology_definition["name"],
            onto_label=ontology_definition["label"],
            onto_comment=ontology_definition.get("comment"),
            project_ontologies=project_ontologies,
            con=con,
            project_remote=project_remote,
            context=context,
            verbose=verbose,
        )
        if not ontology_remote:
            overall_success = False
            continue

        # add the empty resource classes to the remote ontology
        last_modification_date, remote_res_classes, success = _add_resource_classes_to_remote_ontology(
            onto_name=ontology_definition["name"],
            resclass_definitions=ontology_definition.get("resources", []),
            ontology_remote=ontology_remote,
            con=con,
            last_modification_date=ontology_remote.lastModificationDate,
            verbose=verbose,
        )
        if not success:
            overall_success = False

        # add the property classes to the remote ontology
        last_modification_date, success = _add_property_classes_to_remote_ontology(
            onto_name=ontology_definition["name"],
            property_definitions=ontology_definition.get("properties", []),
            ontology_remote=ontology_remote,
            names_and_iris_of_list_nodes=names_and_iris_of_list_nodes,
            con=con,
            last_modification_date=last_modification_date,
            knora_api_prefix=knora_api_prefix,
            verbose=verbose,
        )
        if not success:
            overall_success = False

        # Add cardinalities to class
        success = _add_cardinalities_to_resource_classes(
            resclass_definitions=ontology_definition.get("resources", []),
            ontology_remote=ontology_remote,
            remote_res_classes=remote_res_classes,
            last_modification_date=last_modification_date,
            knora_api_prefix=knora_api_prefix,
            verbose=verbose,
        )
        if not success:
            overall_success = False

    return overall_success


def _create_ontology(
    onto_name: str,
    onto_label: str,
    onto_comment: Optional[str],
    project_ontologies: list[Ontology],
    con: Connection,
    project_remote: Project,
    context: Context,
    verbose: bool,
) -> Optional[Ontology]:
    """
    Create an ontology on the DSP server,
    and add the prefixes defined in the JSON file to its context.
    If the ontology already exists on the DSP server, it is skipped.

    Args:
        onto_name: name of the ontology
        onto_label: label of the ontology
        onto_comment: comment of the ontology
        project_ontologies: ontologies existing on the DSP server
        con: Connection to the DSP server
        project_remote: representation of the project on the DSP server
        context: prefixes and the ontology IRIs they stand for
        verbose: verbose switch

    Raises:
        InputError: if the ontology cannot be created on the DSP server

    Returns:
        representation of the created ontology on the DSP server, or None if it already existed
    """
    # skip if it already exists on the DSP server
    if onto_name in [onto.name for onto in project_ontologies]:
        err_msg = f"Ontology '{onto_name}' already exists on the DSP server. Skipping..."
        print(f"    WARNING: {err_msg}")
        logger.warning(err_msg)
        return None

    print(f"Create ontology '{onto_name}'...")
    logger.info(f"Create ontology '{onto_name}'...")
    ontology_local = Ontology(
        con=con,
        project=project_remote,
        label=onto_label,
        name=onto_name,
        comment=onto_comment,
    )
    try:
        ontology_remote = ontology_local.create()
    except BaseError:
        # if ontology cannot be created, let the error escalate
        logger.exception(f"ERROR while trying to create ontology '{onto_name}'.")
        raise InputError(f"ERROR while trying to create ontology '{onto_name}'.") from None

    if verbose:
        print(f"    Created ontology '{onto_name}'.")
    logger.info(f"Created ontology '{onto_name}'.")

    context.add_context(
        ontology_remote.name,
        ontology_remote.iri + ("" if ontology_remote.iri.endswith("#") else "#"),
    )

    # add the prefixes defined in the JSON file
    for onto_prefix, onto_info in context:
        if onto_info and str(onto_prefix) not in ontology_remote.context:
            onto_iri = onto_info.iri + ("#" if onto_info.hashtag else "")
            ontology_remote.context.add_context(prefix=str(onto_prefix), iri=onto_iri)

    return ontology_remote


def _add_resource_classes_to_remote_ontology(
    onto_name: str,
    resclass_definitions: list[dict[str, Any]],
    ontology_remote: Ontology,
    con: Connection,
    last_modification_date: DateTimeStamp,
    verbose: bool,
) -> tuple[DateTimeStamp, dict[str, ResourceClass], bool]:
    """
    Creates the resource classes (without cardinalities) defined in the "resources" section of an ontology. The
    containing project and the containing ontology must already be existing on the DSP server.
    If an error occurs during creation of a resource class, it is printed out, the process continues, but the success
    status will be false.

    Args:
        onto_name: name of the current ontology
        resclass_definitions: the part of the parsed JSON project file that contains the resources of the current onto
        ontology_remote: representation of the current ontology on the DSP server
        con: connection to the DSP server
        last_modification_date: last modification date of the ontology on the DSP server
        verbose: verbose switch

    Returns:
        last modification date of the ontology,
        new resource classes,
        success status
    """

    overall_success = True
    print("    Create resource classes...")
    logger.info("Create resource classes...")
    new_res_classes: dict[str, ResourceClass] = {}
    sorted_resources = _sort_resources(resclass_definitions, onto_name)
    for res_class in sorted_resources:
        super_classes = res_class["super"]
        if isinstance(super_classes, str):
            super_classes = [super_classes]
        res_class_local = ResourceClass(
            con=con,
            context=ontology_remote.context,
            ontology_id=ontology_remote.iri,
            name=res_class["name"],
            superclasses=super_classes,
            label=LangString(res_class.get("labels")),
            comment=LangString(res_class.get("comments")) if res_class.get("comments") else None,
        )
        try:
            last_modification_date, res_class_remote = res_class_local.create(last_modification_date)
            new_res_classes[str(res_class_remote.iri)] = res_class_remote
            ontology_remote.lastModificationDate = last_modification_date
            if verbose:
                print(f"    Created resource class '{res_class['name']}'")
            logger.info(f"Created resource class '{res_class['name']}'")
        except BaseError:
            err_msg = f"Unable to create resource class '{res_class['name']}'."
            print(f"WARNING: {err_msg}")
            logger.exception(err_msg)
            overall_success = False

    return last_modification_date, new_res_classes, overall_success


def _sort_resources(
    unsorted_resources: list[dict[str, Any]],
    onto_name: str,
) -> list[dict[str, Any]]:
    """
    This method sorts the resource classes in an ontology according to their inheritance order (parent classes first).

    Args:
        unsorted_resources: list of resources from a parsed JSON project file
        onto_name: name of the onto

    Returns:
        sorted list of resource classes
    """

    # do not modify the original unsorted_resources, which points to the original JSON project file
    resources_to_sort = unsorted_resources.copy()
    sorted_resources: list[dict[str, Any]] = []
    ok_resource_names: list[str] = []
    while resources_to_sort:
        # inside the for loop, resources_to_sort is modified, so a copy must be made to iterate over
        for res in resources_to_sort.copy():
            parent_classes = res["super"]
            if isinstance(parent_classes, str):
                parent_classes = [parent_classes]
            parent_classes = [regex.sub(r"^:([^:]+)$", f"{onto_name}:\\1", elem) for elem in parent_classes]
            parent_classes_ok = [not p.startswith(onto_name) or p in ok_resource_names for p in parent_classes]
            if all(parent_classes_ok):
                sorted_resources.append(res)
                res_name = f"{onto_name}:{res['name']}"
                ok_resource_names.append(res_name)
                resources_to_sort.remove(res)
    return sorted_resources


def _add_property_classes_to_remote_ontology(
    onto_name: str,
    property_definitions: list[dict[str, Any]],
    ontology_remote: Ontology,
    names_and_iris_of_list_nodes: dict[str, Any],
    con: Connection,
    last_modification_date: DateTimeStamp,
    knora_api_prefix: str,
    verbose: bool,
) -> tuple[DateTimeStamp, bool]:
    """
    Creates the property classes defined in the "properties" section of an ontology. The
    containing project and the containing ontology must already be existing on the DSP server.
    If an error occurs during creation of a property class, it is printed out, the process continues, but the success
    status will be false.

    Args:
        onto_name: name of the current ontology
        property_definitions: the part of the parsed JSON project file that contains the properties of the current onto
        ontology_remote: representation of the current ontology on the DSP server
        names_and_iris_of_list_nodes: IRIs of list nodes that were already created and are available on the DSP server
        con: connection to the DSP server
        last_modification_date: last modification date of the ontology on the DSP server
        knora_api_prefix: the prefix that stands for the knora-api ontology
        verbose: verbose switch

    Returns:
        a tuple consisting of the last modification date of the ontology, and the success status
    """
    overall_success = True
    print("    Create property classes...")
    logger.info("Create property classes...")
    sorted_prop_classes = _sort_prop_classes(property_definitions, onto_name)
    for prop_class in sorted_prop_classes:
        # get the super-property/ies, valid forms are:
        #   - "prefix:super-property" : fully qualified name of property in another ontology. The prefix has to be
        #     defined in the prefixes part.
        #   - ":super-property" : super-property defined in current ontology
        #   - "super-property" : super-property defined in the knora-api ontology
        #   - if omitted, "knora-api:hasValue" is assumed
        if prop_class.get("super"):
            super_props = []
            for super_class in prop_class["super"]:
                if ":" in super_class:
                    prefix, _class = super_class.split(":")
                    super_props.append(super_class if prefix else f"{ontology_remote.name}:{_class}")
                else:
                    super_props.append(knora_api_prefix + super_class)
        else:
            super_props = ["knora-api:hasValue"]

        # get the "object", valid forms are:
        #   - "prefix:object_name" : fully qualified object. The prefix has to be defined in the prefixes part.
        #   - ":object_name" : The object is defined in the current ontology.
        #   - "object_name" : The object is defined in "knora-api"
        if ":" in prop_class["object"]:
            prefix, _object = prop_class["object"].split(":")
            prop_object = f"{prefix}:{_object}" if prefix else f"{ontology_remote.name}:{_object}"
        else:
            prop_object = knora_api_prefix + prop_class["object"]

        # get the gui_attributes
        gui_attributes = prop_class.get("gui_attributes")
        if gui_attributes and gui_attributes.get("hlist"):
            list_iri = names_and_iris_of_list_nodes[gui_attributes["hlist"]]["id"]
            gui_attributes["hlist"] = f"<{list_iri}>"

        # create the property class
        prop_class_local = PropertyClass(
            con=con,
            context=ontology_remote.context,
            label=LangString(prop_class.get("labels")),
            name=prop_class["name"],
            ontology_id=ontology_remote.iri,
            superproperties=super_props,
            rdf_object=prop_object,
            rdf_subject=prop_class.get("subject"),
            gui_element="salsah-gui:" + prop_class["gui_element"],
            gui_attributes=gui_attributes,
            comment=LangString(prop_class["comments"]) if prop_class.get("comments") else None,
        )
        try:
            last_modification_date, _ = prop_class_local.create(last_modification_date)
            ontology_remote.lastModificationDate = last_modification_date
            if verbose:
                print(f"    Created property class '{prop_class['name']}'")
            logger.info(f"Created property class '{prop_class['name']}'")
        except BaseError:
            err_msg = f"Unable to create property class '{prop_class['name']}'."
            print(f"WARNING: {err_msg}")
            logger.exception(f"Unable to create property class '{prop_class['name']}'.")
            overall_success = False

    return last_modification_date, overall_success


def _sort_prop_classes(
    unsorted_prop_classes: list[dict[str, Any]],
    onto_name: str,
) -> list[dict[str, Any]]:
    """
    In case of inheritance, parent properties must be uploaded before their children. This method sorts the
    properties.

    Args:
        unsorted_prop_classes: list of properties from a parsed JSON project file
        onto_name: name of the onto

    Returns:
        sorted list of properties
    """

    # do not modify the original unsorted_prop_classes, which points to the original JSON project file
    prop_classes_to_sort = unsorted_prop_classes.copy()
    sorted_prop_classes: list[dict[str, Any]] = []
    ok_propclass_names: list[str] = []
    while prop_classes_to_sort:
        # inside the for loop, resources_to_sort is modified, so a copy must be made to iterate over
        for prop in prop_classes_to_sort.copy():
            prop_name = f"{onto_name}:{prop['name']}"
            parent_classes = prop.get("super", "hasValue")
            if isinstance(parent_classes, str):
                parent_classes = [parent_classes]
            parent_classes = [regex.sub(r"^:([^:]+)$", f"{onto_name}:\\1", elem) for elem in parent_classes]
            parent_classes_ok = [not p.startswith(onto_name) or p in ok_propclass_names for p in parent_classes]
            if all(parent_classes_ok):
                sorted_prop_classes.append(prop)
                ok_propclass_names.append(prop_name)
                prop_classes_to_sort.remove(prop)
    return sorted_prop_classes


def _add_cardinalities_to_resource_classes(
    resclass_definitions: list[dict[str, Any]],
    ontology_remote: Ontology,
    remote_res_classes: dict[str, ResourceClass],
    last_modification_date: DateTimeStamp,
    knora_api_prefix: str,
    verbose: bool,
) -> bool:
    """
    Iterates over the resource classes of an ontology of a JSON project definition, and adds the cardinalities to each
    resource class. The resource classes and the properties must already be existing on the DSP server.
    If an error occurs during creation of a cardinality, it is printed out, the process continues, but the success
    status will be false.

    Args:
        resclass_definitions: the part of the parsed JSON project file that contains the resources of the current onto
        ontology_remote: representation of the current ontology on the DSP server
        remote_res_classes: representations of the resource classes on the DSP server
        last_modification_date: last modification date of the ontology on the DSP server
        knora_api_prefix: the prefix that stands for the knora-api ontology
        verbose: verbose switch

    Returns:
        success status
    """
    overall_success = True
    print("    Add cardinalities to resource classes...")
    logger.info("Add cardinalities to resource classes...")
    switcher = {
        "1": Cardinality.C_1,
        "0-1": Cardinality.C_0_1,
        "0-n": Cardinality.C_0_n,
        "1-n": Cardinality.C_1_n,
    }
    for res_class in resclass_definitions:
        res_class_remote = remote_res_classes.get(f"{ontology_remote.iri}#{res_class['name']}")
        if not res_class_remote:
            msg = (
                f"Unable to add cardinalities to resource class '{res_class['name']}': "
                f"This class doesn't exist on the DSP server."
            )
            print(f"WARNINIG: {msg}")
            logger.exception(msg)
            overall_success = False
            continue
        for card_info in res_class.get("cardinalities", []):
            if ":" in card_info["propname"]:
                prefix, prop = card_info["propname"].split(":")
                qualified_propname = card_info["propname"] if prefix else f"{ontology_remote.name}:{prop}"
            else:
                qualified_propname = knora_api_prefix + card_info["propname"]

            try:
                last_modification_date = res_class_remote.addProperty(
                    property_id=qualified_propname,
                    cardinality=switcher[card_info["cardinality"]],
                    gui_order=card_info.get("gui_order"),
                    last_modification_date=last_modification_date,
                )
                if verbose:
                    print(f"    Added cardinality '{card_info['propname']}' to resource class '{res_class['name']}'")
                logger.info(f"Added cardinality '{card_info['propname']}' to resource class '{res_class['name']}'")
            except BaseError:
                err_msg = f"Unable to add cardinality '{qualified_propname}' to resource class {res_class['name']}."
                print(f"WARNING: {err_msg}")
                logger.exception(err_msg)
                overall_success = False

            ontology_remote.lastModificationDate = last_modification_date

    return overall_success



from pathlib import Path

from loguru import logger
from rdflib import Graph

from dsp_tools.commands.validate_data.constants import CARDINALITY_DATA_TTL
from dsp_tools.commands.validate_data.constants import CARDINALITY_REPORT_TTL
from dsp_tools.commands.validate_data.constants import CARDINALITY_SHACL_TTL
from dsp_tools.commands.validate_data.constants import CONTENT_DATA_TTL
from dsp_tools.commands.validate_data.constants import CONTENT_REPORT_TTL
from dsp_tools.commands.validate_data.constants import CONTENT_SHACL_TTL
from dsp_tools.commands.validate_data.models.validation import RDFGraphs
from dsp_tools.commands.validate_data.models.validation import ValidationFilePaths
from dsp_tools.commands.validate_data.models.validation import ValidationReportGraphs
from dsp_tools.commands.validate_data.shacl_cli_validator import ShaclCliValidator
from dsp_tools.commands.validate_data.utils import clean_up_temp_directory
from dsp_tools.commands.validate_data.utils import get_temp_directory
from dsp_tools.error.exceptions import ShaclValidationError


def get_validation_report(
    rdf_graphs: RDFGraphs, shacl_validator: ShaclCliValidator, graph_save_dir: Path | None = None
) -> ValidationReportGraphs:
    tmp_dir = get_temp_directory()
    tmp_path = Path(tmp_dir.name)
    dir_to_save_graphs = graph_save_dir
    try:
        result = _call_shacl_cli(rdf_graphs, shacl_validator, tmp_path)
        return result
    except Exception as e:  # noqa: BLE001
        logger.exception(e)
        dir_to_save_graphs = tmp_path.parent / "validation-graphs"
        msg = (
            f"An error occurred during the data validation. "
            f"Please contact the dsp-tools development team "
            f"with your log files and the files in the directory: {dir_to_save_graphs}"
        )
        raise ShaclValidationError(msg) from None
    finally:
        clean_up_temp_directory(tmp_dir, dir_to_save_graphs)


def _call_shacl_cli(
    rdf_graphs: RDFGraphs, shacl_validator: ShaclCliValidator, tmp_path: Path
) -> ValidationReportGraphs:
    _create_and_write_graphs(rdf_graphs, tmp_path)
    results_graph = Graph()
    conforms = True
    card_files = ValidationFilePaths(
        directory=tmp_path,
        data_file=CARDINALITY_DATA_TTL,
        shacl_file=CARDINALITY_SHACL_TTL,
        report_file=CARDINALITY_REPORT_TTL,
    )
    card_result = shacl_validator.validate(card_files)
    if not card_result.conforms:
        results_graph += card_result.validation_graph
        conforms = False
    content_files = ValidationFilePaths(
        directory=tmp_path,
        data_file=CONTENT_DATA_TTL,
        shacl_file=CONTENT_SHACL_TTL,
        report_file=CONTENT_REPORT_TTL,
    )
    content_result = shacl_validator.validate(content_files)
    if not content_result.conforms:
        results_graph += content_result.validation_graph
        conforms = False
    return ValidationReportGraphs(
        conforms=conforms,
        validation_graph=results_graph,
        shacl_graph=rdf_graphs.cardinality_shapes + rdf_graphs.content_shapes,
        onto_graph=rdf_graphs.ontos + rdf_graphs.knora_api,
        data_graph=rdf_graphs.data,
    )


def _create_and_write_graphs(rdf_graphs: RDFGraphs, tmp_path: Path) -> None:
    logger.debug("Serialise RDF graphs into turtle strings")
    data_str = rdf_graphs.data.serialize(format="ttl")
    ontos_str = rdf_graphs.ontos.serialize(format="ttl")
    card_shape_str = rdf_graphs.cardinality_shapes.serialize(format="ttl")
    content_shape_str = rdf_graphs.content_shapes.serialize(format="ttl")
    knora_api_str = rdf_graphs.knora_api.serialize(format="ttl")
    turtle_paths_and_graphs = [
        (tmp_path / CARDINALITY_DATA_TTL, data_str),
        (tmp_path / CARDINALITY_SHACL_TTL, card_shape_str + ontos_str + knora_api_str),
        (tmp_path / CONTENT_DATA_TTL, data_str + ontos_str + knora_api_str),
        (tmp_path / CONTENT_SHACL_TTL, content_shape_str + ontos_str + knora_api_str),
    ]
    for f_path, content in turtle_paths_and_graphs:
        with open(f_path, "w") as writer:
            writer.write(content)



from dataclasses import dataclass
from dataclasses import field
from typing import Protocol

from dsp_tools.clients.connection import Connection
from dsp_tools.error.exceptions import BaseError
from dsp_tools.error.exceptions import InputError


@dataclass(frozen=True)
class ProjectInfo:
    """Information about a project."""

    project_iri: str


class ProjectClient(Protocol):
    """Interface (protocol) for project-related requests to the DSP-API."""

    con: Connection
    shortcode: str
    project_info: ProjectInfo | None

    def get_project_iri(self) -> str:
        """Get the IRI of the project to which the data is being uploaded."""


@dataclass()
class ProjectClientLive:
    """Client handling project-related requests to the DSP-API."""

    con: Connection
    shortcode: str
    project_info: ProjectInfo | None = field(init=False, default=None)

    def get_project_iri(self) -> str:
        """Get the IRI of the project to which the data is being uploaded."""
        if not self.project_info:
            self.project_info = _get_project_info_from_server(self.con, self.shortcode)
        return self.project_info.project_iri


def _get_project_info_from_server(con: Connection, shortcode: str) -> ProjectInfo:
    project_iri = _get_project_iri_from_server(con, shortcode)
    return ProjectInfo(project_iri=project_iri)


def _get_project_iri_from_server(con: Connection, shortcode: str) -> str:
    url = f"/admin/projects/shortcode/{shortcode}"
    try:
        res = con.get(url)
    except BaseError as e:
        raise InputError(f"A project with shortcode {shortcode} could not be found on the DSP server") from e

    iri: str | None = res.get("project", {}).get("id")
    if not iri:
        raise BaseError(f"Unexpected response from server: {res}")
    return iri



from __future__ import annotations

from lxml import etree

from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.models.permission import PermissionValue
from dsp_tools.error.exceptions import XmlUploadError
from dsp_tools.legacy_models.projectContext import ProjectContext


class XmlPermission:
    """Represents the permission set containing several XmlAllow elements in the XML used for data import"""

    permission_id: str
    _allows: list[XmlAllow]

    def __init__(self, node: etree._Element, project_context: ProjectContext) -> None:
        """
        Constructor which parses a XML DOM permissions element representing an named permission set

        Args:
            node: The DOM node to be processed (representing an a permission set)
            project_context: Context for DOM node traversal
        """
        self._allows = []
        self.permission_id = node.attrib["id"]
        self._allows.extend(XmlAllow(allow_node, project_context) for allow_node in node)

    def get_permission_instance(self) -> Permissions:
        """Returns a list of allow elements of this permission instance"""
        permissions = Permissions()
        for allow in self._allows:
            permissions.add(allow.permission, allow.group)
        return permissions

    def __str__(self) -> str:
        allow_str: list[str] = [f"{allow.permission} {allow.group}" for allow in self._allows]
        return "|".join(allow_str)


class XmlAllow:
    """Represents the allow element of the XML used for data import"""

    _group: str
    _permission: PermissionValue

    def __init__(self, node: etree._Element, project_context: ProjectContext) -> None:
        """
        Constructor which parses the XML DOM allow element

        Args:
            node: The DOM node to be processed (represents a single right in a permission set)
            project_context: Context for DOM node traversal

        Raises:
            XmlUploadError: If an upload fails
        """
        tmp = node.attrib["group"].split(":")

        sysgroups = ["UnknownUser", "KnownUser", "ProjectMember", "Creator", "ProjectAdmin", "SystemAdmin"]
        if len(tmp) > 1:
            if tmp[0]:
                if tmp[0] == "knora-admin" and tmp[1] in sysgroups:
                    self._group = node.attrib["group"]
                else:
                    _group = project_context.group_map.get(node.attrib["group"])
                    if _group is None:
                        raise XmlUploadError(f'Group "{node.attrib["group"]}" is not known: Cannot find project!')
                    self._group = _group
            elif project_context.project_name is None:
                raise XmlUploadError("Project shortname has not been set in ProjectContext")
            else:
                prefixed_custom_groupname = f"{project_context.project_name}:{tmp[1]}"
                if _group := project_context.group_map.get(prefixed_custom_groupname):
                    self._group = _group
        elif tmp[0] in sysgroups:
            self._group = "knora-admin:" + node.attrib["group"]
        else:
            raise XmlUploadError(f'Group "{node.attrib["group"]}" is not known: ')

        if not node.text:
            raise XmlUploadError("No permission set specified")
        self._permission = PermissionValue[node.text]

    @property
    def group(self) -> str:
        """The group specified in the allow element"""
        return self._group

    @property
    def permission(self) -> PermissionValue:
        """The reference to a set of permissions"""
        return self._permission



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from pathlib import Path


@dataclass(frozen=True)
class ServerCredentials:
    """Contains credentials to connect to a server."""

    user: str
    password: str
    server: str
    dsp_ingest_url: str = "http://0.0.0.0:3340"


@dataclass(frozen=True)
class ValidateDataConfig:
    """Contains the configuration for validate data."""

    xml_file: Path
    save_graph_dir: Path | None
    severity: ValidationSeverity
    ignore_duplicate_files_warning: bool
    is_on_prod_server: bool


class ValidationSeverity(Enum):
    ERROR = 3
    WARNING = 2
    INFO = 1



from dataclasses import dataclass
from dataclasses import field
from typing import Any

import pytest

from dsp_tools.commands.xmlupload.project_client import ProjectClientLive
from test.integration.commands.xmlupload.connection_mock import ConnectionMockBase


@dataclass
class ConnectionMock(ConnectionMockBase):
    get_responses: list[dict[str, Any]] = field(default_factory=list)

    def get(
        self,
        route: str,  # noqa: ARG002 (unused-method-argument)
        headers: dict[str, str] | None = None,  # noqa: ARG002 (unused-method-argument)
    ) -> dict[str, Any]:
        return self.get_responses.pop(0)


class TestProjectClientLive:
    """Test the ProjectClientLive class."""

    def test_get_project_iri(self) -> None:
        project_iri_response = {"project": {"id": "http://www.example.org/projects#a"}}
        project_ontologies_response = {"@id": "http://www.example.org/ontologies/a.1/v2"}
        con = ConnectionMock([project_iri_response, project_ontologies_response])
        project_client = ProjectClientLive(con, "")
        project_iri = project_client.get_project_iri()
        expected = "http://www.example.org/projects#a"
        assert project_iri == expected


if __name__ == "__main__":
    pytest.main([__file__])



# mypy: disable-error-code="no-untyped-def"

from pathlib import Path
from uuid import uuid4

from dsp_tools.commands.validate_data.models.validation import ValidationFilePaths
from dsp_tools.commands.validate_data.shacl_cli_validator import ShaclCliValidator

FILE_DIR = Path("testdata/validate-data/shacl_cli")
DATA_TTL = "data.ttl"
SHACL_TTL = "shacl.ttl"


def test_run_validate_cli():
    report_ttl = f"{uuid4()!s}.ttl"
    files = ValidationFilePaths(
        directory=FILE_DIR,
        data_file="data.ttl",
        shacl_file="shacl.ttl",
        report_file=report_ttl,
    )
    validator = ShaclCliValidator()
    validator._run_validate_cli(files)
    assert (FILE_DIR / report_ttl).is_file()
    (FILE_DIR / report_ttl).unlink()


def test_validate_conforms():
    report_ttl = f"{uuid4()!s}.ttl"
    files = ValidationFilePaths(
        directory=FILE_DIR,
        data_file="data.ttl",
        shacl_file="shacl.ttl",
        report_file=report_ttl,
    )
    validator = ShaclCliValidator()
    result = validator.validate(files)
    assert result.conforms
    (FILE_DIR / report_ttl).unlink()



from __future__ import annotations

from dataclasses import dataclass

from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.utils.xml_parsing.models.parsed_resource import KnoraValueType


@dataclass
class ProcessedFileMetadata:
    license_iri: str
    copyright_holder: str
    authorships: list[str]
    permissions: Permissions | None = None


@dataclass
class ProcessedFileValue:
    value: str
    file_type: KnoraValueType
    metadata: ProcessedFileMetadata
    res_id: str
    res_label: str


@dataclass
class ProcessedIIIFUri:
    value: str
    metadata: ProcessedFileMetadata



import json
import urllib.parse
from collections.abc import Iterator
from pathlib import Path
from typing import Any

import pytest
import requests
from pytest_unordered import unordered
from rdflib import RDF
from rdflib import RDFS
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.commands.project.create.project_create_all import create_project
from dsp_tools.commands.xmlupload.xmlupload import xmlupload
from dsp_tools.utils.rdflib_constants import KNORA_API
from test.e2e.setup_testcontainers.ports import ExternalContainerPorts
from test.e2e.setup_testcontainers.setup import get_containers

PROJECT_SHORTCODE = "4125"
E2E_TESTONTO_PREFIX = "e2e-testonto"
SECOND_ONTO_PREFIX = "second-onto"
PROPS_IN_ONTO_JSON = 1
RESCLASSES_IN_ONTO_JSON = 2
USER_IRI_PREFIX = "http://www.knora.org/ontology/knora-admin#"


@pytest.fixture(scope="module")
def container_ports() -> Iterator[ExternalContainerPorts]:
    with get_containers() as metadata:
        yield metadata.ports


@pytest.fixture(scope="module")
def creds(container_ports: ExternalContainerPorts) -> ServerCredentials:
    return ServerCredentials(
        "root@example.com",
        "test",
        f"http://0.0.0.0:{container_ports.api}",
        f"http://0.0.0.0:{container_ports.ingest}",
    )


@pytest.fixture(scope="module")
def e2e_testonto_iri(creds: ServerCredentials) -> str:
    return f"{creds.server}/ontology/{PROJECT_SHORTCODE}/{E2E_TESTONTO_PREFIX}/v2"


@pytest.fixture(scope="module")
def second_onto_iri(creds: ServerCredentials) -> str:
    return f"{creds.server}/ontology/{PROJECT_SHORTCODE}/{SECOND_ONTO_PREFIX}/v2"


@pytest.fixture(scope="module")
def _create_project(creds: ServerCredentials) -> None:
    assert create_project(Path("testdata/json-project/test-project-e2e.json"), creds, verbose=True)


@pytest.fixture(scope="module")
def _xmlupload(_create_project: None, creds: ServerCredentials) -> Iterator[None]:
    assert xmlupload(Path("testdata/xml-data/test-data-e2e.xml"), creds, ".")
    yield
    if found := list(Path.cwd().glob(f"id2iri_{PROJECT_SHORTCODE}_localhost_*.json")):
        found[0].unlink()


@pytest.fixture(scope="module")
def auth_header(_create_project: None, creds: ServerCredentials) -> dict[str, str]:
    payload = {"email": creds.user, "password": creds.password}
    token: str = requests.post(f"{creds.server}/v2/authentication", json=payload, timeout=3).json()["token"]
    return {"Authorization": f"Bearer {token}"}


@pytest.fixture(scope="module")
def project_iri(_create_project: None, creds: ServerCredentials) -> str:
    get_project_route = f"{creds.server}/admin/projects/shortcode/{PROJECT_SHORTCODE}"
    project_iri: str = requests.get(get_project_route, timeout=3).json()["project"]["id"]
    return project_iri


@pytest.mark.usefixtures("_create_project")
def test_project(
    auth_header: dict[str, str], creds: ServerCredentials, e2e_testonto_iri: str, second_onto_iri: str
) -> None:
    get_project_url = f"{creds.server}/admin/projects/shortcode/{PROJECT_SHORTCODE}"
    project = requests.get(get_project_url, headers=auth_header, timeout=3).json()["project"]
    _check_project(project, e2e_testonto_iri, second_onto_iri)

    onto = requests.get(e2e_testonto_iri, headers=auth_header, timeout=3).json()["@graph"]
    _check_props([elem for elem in onto if elem.get("knora-api:isResourceProperty")])
    _check_resclasses([elem for elem in onto if elem.get("knora-api:isResourceClass")])


@pytest.mark.usefixtures("_create_project")
def test_all_get_licenses_enabled(auth_header: dict[str, str], creds: ServerCredentials) -> None:
    response = _get_enabled_licenses(auth_header, creds)
    all_ids = {x["id"] for x in response["data"]}
    assert all_ids == {"http://rdfh.ch/licenses/cc-by-4.0", "http://rdfh.ch/licenses/cc-by-nc-4.0"}


@pytest.mark.usefixtures("_create_project")
def test_default_permissions(creds: ServerCredentials, project_iri: str, auth_header: dict[str, str]) -> None:
    response = requests.get(
        f"{creds.server}/admin/permissions/doap/{urllib.parse.quote_plus(project_iri)}", headers=auth_header, timeout=3
    )
    doaps = response.json()["default_object_access_permissions"]
    assert len(doaps) == 1  # As soon as per-class DOAPs are supported, there will be more

    # The first DOAP is the public one
    assert doaps[0]["forGroup"] == f"{USER_IRI_PREFIX}ProjectMember"
    expected_permissions = [
        {"additionalInformation": f"{USER_IRI_PREFIX}ProjectAdmin", "name": "CR", "permissionCode": 8},
        {"additionalInformation": f"{USER_IRI_PREFIX}ProjectMember", "name": "D", "permissionCode": 7},
        {"additionalInformation": f"{USER_IRI_PREFIX}KnownUser", "name": "V", "permissionCode": 2},
        {"additionalInformation": f"{USER_IRI_PREFIX}UnknownUser", "name": "V", "permissionCode": 2},
    ]
    assert unordered(doaps[0]["hasPermissions"]) == expected_permissions

    # For each class/property which defines their own permissions, there will be a separate DOAP
    pass  # noqa: PIE790


def _get_enabled_licenses(auth_header: dict[str, str], creds: ServerCredentials) -> dict[str, Any]:
    url = (
        f"{creds.server}/admin/projects/shortcode/4125/legal-info/"
        f"licenses?page=1&page-size=25&order=Asc&showOnlyEnabled=true"
    )
    headers = auth_header | {"Accept": "application/json"}
    response = requests.get(url=url, headers=headers, timeout=3).json()
    return dict(response)


@pytest.mark.usefixtures("_xmlupload")
def test_xmlupload(
    auth_header: dict[str, str], project_iri: str, creds: ServerCredentials, e2e_testonto_iri: str
) -> None:
    img_resources = _get_resources(f"{e2e_testonto_iri}#ImageResource", auth_header, project_iri, creds)
    img_graph = Graph()
    img_graph.parse(data=img_resources, format="json-ld")
    _analyze_img_labels(img_graph)
    _analyse_legal_info_of_img_resources(img_graph)
    pdf_resources = _get_resources(f"{e2e_testonto_iri}#PDFResource", auth_header, project_iri, creds)
    _analyze_pdf_resources(pdf_resources)


@pytest.mark.usefixtures("_xmlupload")
def test_all_copyright_holders(auth_header: dict[str, str], creds: ServerCredentials) -> None:
    response = _get_copyright_holders(auth_header, creds)
    assert set(response["data"]) == {
        "DaSCH",
        "Wellcome Collection",
        "Public Domain - Not Protected by Copyright",
        "AI-Generated Content - Not Protected by Copyright",
    }


def _get_copyright_holders(auth_header: dict[str, str], creds: ServerCredentials) -> dict[str, Any]:
    url = f"{creds.server}/admin/projects/shortcode/4125/legal-info/copyright-holders?page=1&page-size=25&order=Asc"
    headers = auth_header | {"Accept": "application/json"}
    response = requests.get(url=url, headers=headers, timeout=3).json()
    return dict(response)


def _check_project(project: dict[str, Any], onto_iri: str, second_onto_iri: str) -> None:
    assert project["shortname"] == "e2e-tp"
    assert project["shortcode"] == PROJECT_SHORTCODE
    assert project["longname"] == "e2e test project"
    assert project["description"] == [{"value": "The e2e test project", "language": "en"}]
    assert project["keywords"] == ["e2e-test-project-keyword"]
    assert project["ontologies"] == [onto_iri, second_onto_iri]


def _check_props(props: list[dict[str, Any]]) -> None:
    assert len(props) == PROPS_IN_ONTO_JSON
    assert props[0]["@id"] == f"{E2E_TESTONTO_PREFIX}:hasText"
    assert props[0]["rdfs:subPropertyOf"] == {"@id": "knora-api:hasValue"}
    assert props[0]["knora-api:objectType"] == {"@id": "knora-api:TextValue"}
    assert props[0]["rdfs:label"] == "Text"


def _check_resclasses(resclasses: list[dict[str, Any]]) -> None:
    assert len(resclasses) == RESCLASSES_IN_ONTO_JSON
    res_1, res_2 = resclasses

    assert res_1["@id"] == f"{E2E_TESTONTO_PREFIX}:ImageResource"
    assert res_1["rdfs:label"] == "Image Resource"
    cards_1 = res_1["rdfs:subClassOf"]
    fileval_card = [x for x in cards_1 if x.get("owl:onProperty", {}).get("@id") == "knora-api:hasStillImageFileValue"]
    assert len(fileval_card) == 1
    assert fileval_card[0].get("owl:cardinality") == 1
    hasText_card = [x for x in cards_1 if x.get("owl:onProperty", {}).get("@id") == f"{E2E_TESTONTO_PREFIX}:hasText"]
    assert len(hasText_card) == 1

    assert res_2["@id"] == f"{E2E_TESTONTO_PREFIX}:PDFResource"
    assert res_2["rdfs:label"] == "PDF Resource"


def _get_resources(resclass_iri: str, auth_header: dict[str, str], project_iri: str, creds: ServerCredentials) -> str:
    resclass_iri_encoded = urllib.parse.quote_plus(resclass_iri)
    get_resources_route = f"{creds.server}/v2/resources?resourceClass={resclass_iri_encoded}&page=0"
    headers = auth_header | {"X-Knora-Accept-Project": project_iri}
    response = requests.get(get_resources_route, timeout=3, headers=headers).json()
    resources = json.dumps(response)
    return resources


def _analyze_img_labels(img_graph: Graph) -> None:
    labels = {str(x) for x in img_graph.objects(predicate=RDFS.label)}
    assert labels == {
        "Resource 1",
        "Resource 2",
        "ImageResource_uses_doap_should_be_preview",
        "ImageResource_overrides_doap_is_open",
    }


def _analyse_legal_info_of_img_resources(img_graph: Graph) -> None:
    open_permissions = Literal(
        "CR knora-admin:ProjectAdmin|D knora-admin:ProjectMember|V knora-admin:KnownUser,knora-admin:UnknownUser"
    )
    res_1_iri = next(img_graph.subjects(RDFS.label, Literal("Resource 1")))
    assert next(img_graph.objects(res_1_iri, KNORA_API.hasPermissions)) == open_permissions
    file_1_iri = next(img_graph.objects(res_1_iri, KNORA_API.hasStillImageFileValue))
    assert next(img_graph.objects(file_1_iri, RDF.type)) == KNORA_API.StillImageFileValue
    assert next(img_graph.objects(file_1_iri, KNORA_API.hasAuthorship)) == Literal("Johannes Nussbaum")
    assert next(img_graph.objects(file_1_iri, KNORA_API.hasCopyrightHolder)) == Literal("DaSCH")
    assert next(img_graph.objects(file_1_iri, KNORA_API.hasLicense)) == URIRef("http://rdfh.ch/licenses/cc-by-4.0")
    res_2_iri = next(img_graph.subjects(RDFS.label, Literal("Resource 2")))
    file_2_iri = next(img_graph.objects(res_2_iri, KNORA_API.hasStillImageFileValue))
    assert next(img_graph.objects(file_2_iri, RDF.type)) == KNORA_API.StillImageExternalFileValue
    assert next(img_graph.objects(file_2_iri, KNORA_API.hasPermissions)) == open_permissions
    assert next(img_graph.objects(file_2_iri, KNORA_API.hasAuthorship)) == Literal("Cavanagh, Annie")
    assert next(img_graph.objects(file_2_iri, KNORA_API.hasCopyrightHolder)) == Literal("Wellcome Collection")
    assert next(img_graph.objects(file_2_iri, KNORA_API.hasLicense)) == URIRef("http://rdfh.ch/licenses/cc-by-nc-4.0")
    val_with_comment = next(img_graph.subjects(KNORA_API.valueAsString, Literal("first text value")))
    assert next(img_graph.objects(val_with_comment, KNORA_API.valueHasComment)) == Literal("Comment")
    val_with_perm = next(img_graph.subjects(KNORA_API.valueAsString, Literal("second text value")))
    assert next(img_graph.objects(val_with_perm, KNORA_API.hasPermissions)) == open_permissions


def _analyze_pdf_resources(pdf_resources: str) -> None:
    g = Graph()
    g.parse(data=pdf_resources, format="json-ld")
    res_iri = next(g.subjects(RDFS.label, Literal("Resource 3")))
    file_iri = next(g.objects(res_iri, KNORA_API.hasDocumentFileValue))
    open_permissions = Literal(
        "CR knora-admin:ProjectAdmin|D knora-admin:ProjectMember|V knora-admin:KnownUser,knora-admin:UnknownUser"
    )
    assert next(g.objects(file_iri, RDF.type)) == KNORA_API.DocumentFileValue
    assert next(g.objects(file_iri, KNORA_API.hasAuthorship)) == Literal("Nora Ammann")
    assert next(g.objects(file_iri, KNORA_API.hasCopyrightHolder)) == Literal("DaSCH")
    assert next(g.objects(file_iri, KNORA_API.hasLicense)) == URIRef("http://rdfh.ch/licenses/cc-by-4.0")
    assert next(g.objects(file_iri, KNORA_API.hasPermissions)) == open_permissions



from __future__ import annotations

import os

from dotenv import load_dotenv
from lxml import etree

from dsp_tools.error.xmllib_warnings import MessageInfo
from dsp_tools.error.xmllib_warnings_util import emit_xmllib_input_warning
from dsp_tools.error.xmllib_warnings_util import raise_input_error
from dsp_tools.xmllib.internal.constants import DASCH_SCHEMA
from dsp_tools.xmllib.internal.constants import XML_NAMESPACE_MAP
from dsp_tools.xmllib.internal.serialise_file_value import serialise_file_value
from dsp_tools.xmllib.internal.serialise_values import serialise_values
from dsp_tools.xmllib.internal.type_aliases import AnyResource
from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.models.dsp_base_resources import AudioSegmentResource
from dsp_tools.xmllib.models.dsp_base_resources import LinkResource
from dsp_tools.xmllib.models.dsp_base_resources import RegionResource
from dsp_tools.xmllib.models.dsp_base_resources import VideoSegmentResource
from dsp_tools.xmllib.models.internal.file_values import AuthorshipLookup
from dsp_tools.xmllib.models.internal.values import ColorValue
from dsp_tools.xmllib.models.internal.values import LinkValue
from dsp_tools.xmllib.models.internal.values import Richtext
from dsp_tools.xmllib.models.internal.values import Value
from dsp_tools.xmllib.models.res import Resource

load_dotenv()


def serialise_resources(resources: list[AnyResource], authorship_lookup: AuthorshipLookup) -> list[etree._Element]:
    """
    Serialise all the resources

    Args:
        resources: list of resources
        authorship_lookup: lookup to map the authors to the corresponding IDs

    Returns:
        serialised resources
    """
    env_var = str(os.getenv("XMLLIB_SORT_RESOURCES")).lower()
    if env_var == "true":
        resources = sorted(resources, key=lambda x: x.res_id)
    return [_serialise_one_resource(x, authorship_lookup) for x in resources]


def _serialise_one_resource(res: AnyResource, authorship_lookup: AuthorshipLookup) -> etree._Element:
    match res:
        case Resource():
            return _serialise_generic_resource(res, authorship_lookup)
        case RegionResource():
            return _serialise_region(res)
        case LinkResource():
            return _serialise_link(res)
        case AudioSegmentResource():
            return _serialise_segment(res, "audio-segment")
        case VideoSegmentResource():
            return _serialise_segment(res, "video-segment")
        case _:
            raise_input_error(
                MessageInfo(
                    f"An unknown resource was added to the root. "
                    f"Only Resource, RegionResource, LinkResource, VideoSegmentResource, AudioSegmentResource "
                    f"are permitted."
                    f"The input type is {res.__class__.__name__}"
                )
            )


def _serialise_generic_resource(res: Resource, authorship_lookup: AuthorshipLookup) -> etree._Element:
    ele = _make_generic_resource_element(res, "resource")
    ele.attrib["restype"] = res.restype
    if res.file_value:
        auth_id = authorship_lookup.get_id(res.file_value.metadata.authorship)
        ele.append(serialise_file_value(res.file_value, auth_id))
    ele.extend(serialise_values(res.values))
    return ele


def _serialise_region(res: RegionResource) -> etree._Element:
    ele = _make_generic_resource_element(res, "region")
    ele.extend(_serialise_geometry_shape(res))
    ele.extend(serialise_values(res.values))
    return ele


def _serialise_geometry_shape(res: RegionResource) -> list[etree._Element]:
    prop_list: list[etree._Element] = []
    if not res.geometry:
        emit_xmllib_input_warning(
            MessageInfo(
                "The region resource does not have a geometry. Please note that an xmlupload will fail.", res.res_id
            )
        )
        return prop_list
    geo_prop = etree.Element(f"{DASCH_SCHEMA}geometry-prop", name="hasGeometry", nsmap=XML_NAMESPACE_MAP)
    geo_attrib = (
        {"permissions": res.permissions.value} if res.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS else {}
    )
    ele = etree.Element(f"{DASCH_SCHEMA}geometry", nsmap=XML_NAMESPACE_MAP, attrib=geo_attrib)
    ele.text = res.geometry.to_json_string()
    geo_prop.append(ele)
    prop_list.append(geo_prop)
    prop_list.extend(
        serialise_values([ColorValue(value=res.geometry.color, prop_name="hasColor", permissions=res.permissions)]),
    )
    return prop_list


def _serialise_link(res: LinkResource) -> etree._Element:
    problems = []
    if not any([isinstance(x, Richtext) for x in res.values]):
        problems.append("at least one comment")
    if not any([isinstance(x, LinkValue) for x in res.values]):
        problems.append("at least two links")
    if problems:
        msg = f"A link object requires {' and '.join(problems)}. Please note that an xmlupload will fail."
        emit_xmllib_input_warning(MessageInfo(msg, res.res_id))
    ele = _make_generic_resource_element(res, "link")
    ele.extend(serialise_values(res.values))
    return ele


def _serialise_segment(res: AudioSegmentResource | VideoSegmentResource, segment_type: str) -> etree._Element:
    seg = _make_generic_resource_element(res, segment_type)
    seg.extend(_serialise_segment_children(res))
    return seg


def _make_generic_resource_element(res: AnyResource, res_type: str) -> etree._Element:
    attribs = {"label": res.label, "id": res.res_id}
    if res.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
        attribs["permissions"] = res.permissions.value
    return etree.Element(f"{DASCH_SCHEMA}{res_type}", attrib=attribs, nsmap=XML_NAMESPACE_MAP)


def _serialise_segment_children(segment: AudioSegmentResource | VideoSegmentResource) -> list[etree._Element]:
    # The segment elements need to be in a specific order in order to pass the XSD validation
    # Therefore, this filtering according to the properties is necessary
    segment_elements = _serialise_according_to_prop_name(segment.values, "isSegmentOf")
    bounds_attrib = {
        "segment_start": str(segment.segment_bounds.segment_start),
        "segment_end": str(segment.segment_bounds.segment_end),
    }
    if segment.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
        bounds_attrib["permissions"] = segment.permissions.value
    segment_elements.append(
        etree.Element(
            f"{DASCH_SCHEMA}hasSegmentBounds",
            attrib=bounds_attrib,
            nsmap=XML_NAMESPACE_MAP,
        )
    )
    segment_elements.extend(_serialise_according_to_prop_name(segment.values, "hasTitle"))
    segment_elements.extend(_serialise_according_to_prop_name(segment.values, "hasComment"))
    segment_elements.extend(_serialise_according_to_prop_name(segment.values, "hasDescription"))
    segment_elements.extend(_serialise_according_to_prop_name(segment.values, "hasKeyword"))
    segment_elements.extend(_serialise_according_to_prop_name(segment.values, "relatesTo"))
    return segment_elements


def _serialise_according_to_prop_name(values: list[Value], prop_name: str) -> list[etree._Element]:
    to_serialise = (x for x in values if x.prop_name == prop_name)
    return [_make_element_with_text(x) for x in to_serialise]


def _make_element_with_text(value: Value) -> etree._Element:
    attribs = {}
    if value.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
        attribs["permissions"] = value.permissions.value
    if value.comment is not None:
        attribs["comment"] = str(value.comment)
    ele = etree.Element(f"{DASCH_SCHEMA}{value.prop_name}", nsmap=XML_NAMESPACE_MAP, attrib=attribs)
    ele.text = value.value
    return ele



from dsp_tools.xmllib import LicenseRecommended
from dsp_tools.xmllib.models.dsp_base_resources import RegionResource
from dsp_tools.xmllib.models.res import Resource
from dsp_tools.xmllib.models.root import XMLRoot
from dsp_tools.xmllib.value_checkers import is_bool_like
from dsp_tools.xmllib.value_converters import convert_to_bool_string


def test_xmllib() -> None:
    """
    This does NOT test the xmllib.
    This only tests if all dependencies used in the xmllib are shipped in the distribution.
    For this purpose, it is enough to import all modules of the xmllib.
    Indirect imports are okay, too, i.e. modules that are imported by modules that are imported here.
    """
    root = XMLRoot("0000", "my_onto")
    my_res = Resource.create_new("my_res", "restype", "label")
    if is_bool_like("0"):
        my_res.add_bool("my_prop", convert_to_bool_string("0"))
    my_res.add_file("my_file", LicenseRecommended.CC.BY, "copy", ["auth"])
    region = RegionResource.create_new("res_id", "label", "region_of")
    root.add_resource(region)
    root.add_resource(my_res)



import pickle
from pathlib import Path

from dsp_tools.commands.xmlupload.iri_resolver import IriResolver
from dsp_tools.commands.xmlupload.models.processed.res import ProcessedResource
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedSimpleText
from dsp_tools.commands.xmlupload.models.upload_state import UploadState
from dsp_tools.commands.xmlupload.upload_config import DiagnosticsConfig
from dsp_tools.commands.xmlupload.upload_config import UploadConfig
from dsp_tools.commands.xmlupload.xmlupload import _save_upload_state


def test_save_upload_state(tmp_path: Path) -> None:
    save_location = tmp_path / "upload_state.pkl"
    config = UploadConfig(diagnostics=DiagnosticsConfig(save_location=save_location))
    upload_state = UploadState(
        pending_resources=[
            ProcessedResource("id", "type", "label", None, [ProcessedSimpleText("Some text", "prop", None, None)])
        ],
        failed_uploads=[],
        iri_resolver=IriResolver({"foo": "bar"}),
        pending_stash=None,
        config=config,
    )
    msg = _save_upload_state(upload_state)
    with open(save_location, "rb") as f:
        saved_state: UploadState = pickle.load(f)  # noqa: S301 (deserialization of untrusted data)
    assert msg == f"Saved the current upload state to {save_location}.\n"
    assert len(upload_state.pending_resources) == len(saved_state.pending_resources)
    assert [r.res_id for r in upload_state.pending_resources] == [r.res_id for r in saved_state.pending_resources]
    assert upload_state.iri_resolver.lookup == saved_state.iri_resolver.lookup
    assert upload_state.pending_stash == saved_state.pending_stash
    assert upload_state.config == saved_state.config
